
#include <viz_system.h>

static int checksum(
const uint8_t *s
)
{
    uint8_t sum1= ((uint8_t)'G')^((uint8_t)'P');
    while(1)
    {
	uint8_t c= *(s++);
	if(c==0 || c=='*') break;
	sum1^= c;
    }

    uint32_t sum2= 0;
    while(1)
    {
	int c= *(s++);
	c-= (c>'9' ? 'A'-10 : '0');
	if(c<0 || c>15) break;
	sum2<<= 4;
	sum2+= c;
    }

    return sum2!=(uint32_t)sum1;
}

static int gll=		1;
static int gga=		1;
static int rmc=		1;
static int gsa=		1;
static int rmb=		1;
static int apb=		1;
static int gsv=		1;
static int fix=		0;
static int verbose=	0;
static int dump=	0;

static int parseNMEAAngle(
const uint8_t *angle,
double *result
)
{
    const uint8_t *p= angle;
    while(1)
    {
	if(p[0]=='.') break;
	else if(p[0]<'0' || p[0]>'9') return -1;
	++p;
    }

    p-=2;
    if(p<angle) return -1;
    int min= (p[0]-'0')*10 + (p[1]-'0');

    int deg= 0;
    const uint8_t *s= angle;
    while(s<p)
    {
	deg*= 10;
	deg+= (s[0]-'0');
	++s;
    }

    int div= 1;
    const uint8_t *e= p+3;
    double fracMin= 0.0;
    while(e[0])
    {
	div*= 10;
	fracMin*=10.0;
	fracMin+= e[0]-'0';
	++e;
    }

    double dMin= ((double)min) + (fracMin/(double)div);
    result[0]= ((double)deg) + dMin/60.0;
    return 0;
}

static void dumpFix(
const uint8_t *latitude,
const uint8_t *latitudeSign,
const uint8_t *longitude,
const uint8_t *longitudeSign,
const uint8_t *utc,
const uint8_t *altitude,
const uint8_t *nSat
)
{
    if(fix)
    {
	double la;
	double lo;
	if(parseNMEAAngle(latitude,&la)<0) return;
	if(parseNMEAAngle(longitude,&lo)<0) return;
	if(latitudeSign[0]!='N') la= -la;
	if(longitudeSign[0]!='E') lo= -lo;

	system::error(
	    0,
	    "%g %g %s %c%ch %c%cm %ss %s\t\t(Latitude Longitude Altitude UTC NumberOfSatellitesUsedForFix)\n",
	    la,
	    lo,
	    altitude,
	    utc[0],
	    utc[1],
	    utc[2],
	    utc[3],
	    utc+4,
	    nSat
	);
    }
}

static void sentence(
const uint8_t *s
)
{
    if(verbose) system::error(0,"Raw GPS sentence: %s\n", s);
    if(checksum(s))
    {
	if(verbose) system::error(0,"\tIncorrect checksum, sentence discarded.\n");
	return;
    }

    int nFields= 0;
    const uint8_t *src= s;
    uint8_t fields[81][81];
    uint8_t *dst= fields[1];
    while(1)
    {
	int c= *(src++);
	if(c!=0 && c!=',' && c!='*') *(dst++)= c;
	else
	{
	    dst[0]= 0;
	    if(c==0) break;
	    else dst= fields[1+(nFields++)];
	}
    }

    if(s[0]=='G' && s[1]=='L' && s[2]=='L')
    {
	if(gll)
	{
	    if(verbose) system::error(0,"\tLatitude:%s %s\n",fields[1],fields[2]);
	    if(verbose) system::error(0,"\tLongitude:%s %s\n",fields[3],fields[4]);
	    if(verbose) system::error(0,"\tUTC @ fix (hhmmss.ss): %s\n",fields[5]);
	    if(verbose) system::error(0,"\tFix status (A=Valid V=Invalid): %s\n",fields[6]);
	}
    }
    else if(s[0]=='G' && s[1]=='G' && s[2]=='A')
    {
	if(gga)
	{
	    if(verbose) system::error(0,"\tUTC @ fix (hhmmss.ss): %s\n",fields[1]);
	    if(verbose) system::error(0,"\tLatitude:%s %s\n",fields[2],fields[3]);
	    if(verbose) system::error(0,"\tLongitude:%s %s\n",fields[4],fields[5]);
	    if(verbose) system::error(0,"\tGPS signal quality (0=Bad, 1=GPS, 2=DGPS): %s\n", fields[6]);
	    if(verbose) system::error(0,"\tNumber of satellites in view: %s\n", fields[7]);
	    if(verbose) system::error(0,"\tHorizontal Dilution of precision: %s\n", fields[8]);
	    if(verbose) system::error(0,"\tAntenna Altitude above/below mean-sea-level (geoid): %s\n", fields[9]);
	    if(verbose) system::error(0,"\tUnits of antenna altitude, meters: %s\n",  fields[10]);
	    if(verbose) system::error(0,"\tGeoidal separation: %s\n",  fields[11]);
	    if(verbose) system::error(0,"\tUnits of geoidal separation, meters: %s\n",  fields[12]);
	    if(verbose) system::error(0,"\tAge of differential GPS data, in seconds since last SC104 type 1 or 9 update: %s\n", fields[13]);
	    if(verbose) system::error(0,"\tDifferential reference station ID: %s\n",  fields[14]);
	    if(fields[6][0]=='1') dumpFix(fields[2],fields[3],fields[4],fields[5],fields[1],fields[9],fields[7]);
	}
    }
    else if(s[0]=='R' && s[1]=='M' && s[2]=='C')
    {
	if(rmc)
	{
	    if(verbose) system::error(0,"\tUTC @ fix (hhmmss.ss): %s\n",fields[1]);
	    if(verbose) system::error(0,"\tFix status (A=Valid V=Invalid): %s\n",fields[2]);
	    if(verbose) system::error(0,"\tLatitude:%s %s\n",fields[3],fields[4]);
	    if(verbose) system::error(0,"\tLongitude:%s %s\n",fields[5],fields[6]);
	    if(verbose) system::error(0,"\tSpeed over ground, knots: %s\n", fields[7]);
	    if(verbose) system::error(0,"\tTrack made good, degrees true: %s\n", fields[8]);
	    if(verbose) system::error(0,"\tDate, ddmmyy: %s\n", fields[9]);
	    if(verbose) system::error(0,"\tMagnetic Variation, degrees: %s %s\n", fields[10], fields[11]);
	}
    }
    else if(s[0]=='G' && s[1]=='S' && s[2]=='A')
    {
	if(gsa)
	{
	    if(verbose) system::error(0,"\tMode (Manual/Auto): %s\n",fields[1]);
	    if(verbose) system::error(0,"\tFix (1=Available,2=2D,3=3D): %s\n",fields[2]);
	    if(verbose) system::error(0,"\tSatellites used for solution:\n");
	    if(fields[ 3][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 3]);
	    if(fields[ 4][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 4]);
	    if(fields[ 5][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 5]);
	    if(fields[ 6][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 6]);
	    if(fields[ 7][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 7]);
	    if(fields[ 8][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 8]);
	    if(fields[ 9][0])	if(verbose) system::error(0,"\t\t%s\n",fields[ 9]);
	    if(fields[10][0])	if(verbose) system::error(0,"\t\t%s\n",fields[10]);
	    if(fields[11][0])	if(verbose) system::error(0,"\t\t%s\n",fields[11]);
	    if(fields[12][0])	if(verbose) system::error(0,"\t\t%s\n",fields[12]);
	    if(fields[13][0])	if(verbose) system::error(0,"\t\t%s\n",fields[13]);
	    if(fields[14][0])	if(verbose) system::error(0,"\t\t%s\n",fields[14]);
	    if(verbose) system::error(0,"\tP dilution of precision: %s\n",fields[15]);
	    if(verbose) system::error(0,"\tH dilution of precision: %s\n",fields[16]);
	    if(verbose) system::error(0,"\tV dilution of precision: %s\n",fields[17]);
	}
    }
    else if(s[0]=='R' && s[1]=='M' && s[2]=='B')
    {
	if(rmb)
	{
	    if(verbose) system::error(0,"\tData status (A=Valid,V=Invalid): %s\n",fields[1]);
	    if(verbose) system::error(0,"\tCross track error (XTE), miles and direction to steer(L or R):%s %s\n",fields[2],fields[3]);
	    if(verbose) system::error(0,"\tOrigin waypoint ID: %s\n",fields[4]);
	    if(verbose) system::error(0,"\tDestination waypoint ID: %s\n",fields[5]);
	    if(verbose) system::error(0,"\tDestination waypoint latitude: %s %s\n",fields[6],fields[7]);
	    if(verbose) system::error(0,"\tDestination waypoint longitude: %s %s\n",fields[8],fields[9]);
	    if(verbose) system::error(0,"\tRange in naut. miles from present fix to destination waypoint (great circle): %s\n",fields[10]);
	    if(verbose) system::error(0,"\tBearing, True, Great Circle, present fix to destination waypoint: %s\n",fields[11]);
	    if(verbose) system::error(0,"\tClosing velocity to destination (knots): %s\n", fields[12]);
	    if(verbose) system::error(0,"\tArrival: %s\n", fields[13]);
	}
    }
    else if(s[0]=='A' && s[1]=='P' && s[2]=='B')
    {
	if(apb)
	{
	    if(verbose) system::error(0,"\tMessage status 1 (A=Valid,V=Warning): %s\n",fields[1]);
	    if(verbose) system::error(0,"\tMessage status 2 (A=Valid,V=Warning): %s\n",fields[2]);
	    if(verbose) system::error(0,"\tMagnitude of cross-track error: %s\n",fields[3]);
	    if(verbose) system::error(0,"\tDirection to steer to (L=Left, R=Right): %s\n",fields[4]);
	    if(verbose) system::error(0,"\tXTE units (nautical miles): %s\n",fields[5]);
	    if(verbose) system::error(0,"\tStatus (A=Arrival circle entered): %s\n",fields[6]);
	    if(verbose) system::error(0,"\tStatus (A=Perpendicular passed at waypoint): %s\n",fields[7]);
	    if(verbose) system::error(0,"\tBearing, origin to destination (Magnetic or true): %s %s\n",fields[8],fields[9]);
	    if(verbose) system::error(0,"\tDestination waypoint ID: %s\n",fields[10]);
	    if(verbose) system::error(0,"\tBearing, present position to destination (Magnetic or true): %s %s\n",fields[11],fields[12]);
	    if(verbose) system::error(0,"\theading to steer to destination waypoint (Magnetic or true): %s %s\n",fields[13],fields[14]);
	}
    }
    else if(s[0]=='G' && s[1]=='S' && s[2]=='V')
    {
	if(gsv)
	{
	    if(verbose) system::error(0,"\tTotal number of messages: %s\n",fields[1]);
	    if(verbose) system::error(0,"\tMessage number: %s\n",fields[2]);
	    if(verbose) system::error(0,"\tTotal number of satellites in view: %s\n",fields[3]);
	    if(nFields>4 && fields[4][0])
	    {
		if(verbose) system::error(0,"\tSatellite PRN #: %s\n",fields[4]);
		if(verbose) system::error(0,"\tElevation (degrees): %s\n",fields[5]);
		if(verbose) system::error(0,"\tAzimuth (degrees true): %s\n",fields[6]);
		if(verbose) system::error(0,"\tSignal to noise ratio (dB) 0 if not tracking: %s\n",fields[7]);
	    }
	    if(nFields>8  && fields[8][0])
	    {
		if(verbose) system::error(0,"\tSatellite PRN #: %s\n",fields[8]);
		if(verbose) system::error(0,"\tElevation (degrees): %s\n",fields[9]);
		if(verbose) system::error(0,"\tAzimuth (degrees true): %s\n",fields[10]);
		if(verbose) system::error(0,"\tSignal to noise ratio (dB) 0 if not tracking: %s\n",fields[11]);
	    }
	    if(nFields>12  && fields[12][0])
	    {
		if(verbose) system::error(0,"\tSatellite PRN #: %s\n",fields[12]);
		if(verbose) system::error(0,"\tElevation (degrees): %s\n",fields[13]);
		if(verbose) system::error(0,"\tAzimuth (degrees true): %s\n",fields[14]);
		if(verbose) system::error(0,"\tSignal to noise ratio (dB) 0 if not tracking: %s\n",fields[15]);
	    }
	    if(nFields>16  && fields[16][0])
	    {
		if(verbose) system::error(0,"\tSatellite PRN #: %s\n",fields[16]);
		if(verbose) system::error(0,"\tElevation (degrees): %s\n",fields[17]);
		if(verbose) system::error(0,"\tAzimuth (degrees true): %s\n",fields[18]);
		if(verbose) system::error(0,"\tSignal to noise ratio (dB) 0 if not tracking: %s\n",fields[19]);
	    }
	}
    }
    else
    {
	if(verbose) system::error(0,"Unknown GPS sentence %s\n",s);
    }
}

static void input(
uint8_t c
)
{
    static uint8_t *p;
    static int state= 0;
    static uint8_t buffer[85];
    if(dump) system::error(0,"%c",c);
    switch(state)
    {
	case 0:
	    if(c=='\n')	state= 1;
	    break;
	case 1:
	    if(c=='$')	state= 2;
	    else	state= 0;
	    break;
	case 2:
	    if(c=='G')	state= 3;
	    else	state= 0;
	    break;
	case 3:
	    if(c=='P')	state= 4;
	    else	state= 0;
	    p= buffer;
	    break;
	case 4:
	    if(c=='\r')
	    {
		p[0]= 0;
		state= 0;
		sentence(buffer);
	    }
	    else
	    {
		*(p++)= c;
		if(p-buffer>=80) state= 0;
	    }
	    break;
    }
}

static void usage(
const char *msg= 0
)
{
    if(msg) system::error(0,"Error:%s\n\n",msg);

    system::error(0,"Usage: gps [-s server] [-p port] [-d] [-v]\n\n");
    system::error(0,"\t-s Server to connect to\n"); 
    system::error(0,"\t-p Port to connect to\n"); 
    system::error(0,"\t-c Output current fix only.\n"); 
    system::error(0,"\t-d Raw dump of device output\n"); 
    system::error(0,"\t-v Convert NMEA 2.0 to human readable form\n\n"); 

    system::exit(1);
}

int main(
int	argc,
char	**argv
)
{
    int port= 2222;
    const char *server= "localhost";

    while(*++argv)
    {
	if(argv[0][0]=='-' && argv[0][1]=='s' && argv[0][2]==0)
	{
	    if(argv[1] && argv[1][0])	server= argv[1];
	    else			usage("Incorrect server name");
	    ++argv;
	}
	else if(argv[0][0]=='-' && argv[0][1]=='p' && argv[0][2]==0)
	{
	    if(argv[1]==0 || 1!=system::sscanf(argv[1],"%d",&port)) usage("Incorrect port");
	    ++argv;
	}
	else if(argv[0][0]=='-' && argv[0][1]=='d' && argv[0][2]==0)
	{
	    ++dump;
	}
	else if(argv[0][0]=='-' && argv[0][1]=='v' && argv[0][2]==0)
	{
	    ++verbose;
	}
	else if(argv[0][0]=='-' && argv[0][1]=='c' && argv[0][2]==0)
	{
	    ++fix;
	}
	else
	{
	    usage();
	}
    }

    system::error(0,"GPS3D V1.20  - gps - Please do not use for navigation.\n");
    if(fix==0 && dump==0 && verbose==0) fix= 1;

    void *socket= system::makeSocket(server,port);
    if(socket==(void*)-1) system::error(1,"fatal: can't connect to GPS server %s:%d\n",server,port);

    while(1)
    {
	uint8_t c;
	int r= system::readSocket(socket,&c,1);
	if(r==1) input(c);
	else if(system::lastError()!=system::eAgain && system::lastError()!=system::eInterrupt) system::error(1,"lost connection to server\n");
    }

    return 0;
}

