
#include <viz_system.h>

class Client
{
public:
	void	*socket;
	char	*info;

	Client	*next;
	Client	*prev;

	int	state;
	uint8_t	*input;
	uint8_t	*p;
};

static int	    readOnly;
static int	    clientCnt;
static Client	    *clients;
static const char   *logFile;

static void log(
const char *msg
)
{
    static int logFD= -1;
    if(logFile==0) return;
    if(logFD<0) logFD= system::open(logFile,system::writeOnly|system::creatFlag|system::truncFlag,0777);
    if(logFD>=0) system::write(logFD,msg,system::strlen(msg));
}

static void addClient(
void	    *s,
const char  *info
)
{
    Client *client= new Client;
    client->socket= s;
    client->info= system::strdup(info);
    client->prev= 0;
    client->next= clients;
    client->state= 0;
    client->input= 0;
    client->p= 0;
    clients= client;
    ++clientCnt;

    char buf[1024];
    system::sprintf(buf,"Incoming connection: %s\n", client->info);
    log(buf);
}

static void removeClient(
Client *c
)
{
    char buf[1024];
    system::sprintf(buf,"Closing connection: %s\n", c->info);
    log(buf);

    if(c->next) c->next->prev= c->prev;
    if(c->prev) c->prev->next= c->next;
    else clients= c->next;

    if(c->input) system::free(c->input);
    system::closeSocket(c->socket);
    system::free(c->info);
    delete c;

    --clientCnt;
}

static void clientInput(
	void	*gps,
	Client	*client,
const	uint8_t	*input,
	int	size
)
{
    if(readOnly)
    {
	const char *message= "Access to GPS device unauthorized\r\n";
	int n= system::strlen(message);
	while(n>0)
	{
	    int w= system::writeSocket(client->socket,message,n);
	    if(w>=0) n-= w;
	    else if(
		system::lastError()!=system::eAgain	&&
		system::lastError()!=system::eInterrupt
	    )
	    {
		removeClient(client);
		n= -1;
	    }
	}
    }

    if(client->input==0)
    {
	client->input= (uint8_t*)system::malloc(82);
	client->p= client->input;
	client->state= 0;
    }

    while(size--)
    {
	int c= *(input++);
	if(client->state==0 && c=='$') client->state= 1;
	if(client->state==1)
	{
	    if(c=='\r')
	    {
		*(client->p++)= '\r';
		*(client->p++)= '\n';
		client->p[0]= 0;

		const uint8_t *p= client->input;
		int n= client->p-client->input;
		while(n>0)
		{
		    int r= system::writeSerial(gps,p,n);
		    if(r>=0) n-= r;
		    else if(system::lastError()!=system::eAgain && system::lastError()!=system::eInterrupt) break;
		}

		client->p= client->input;
		client->state= 0;
	    }
	    else
	    {
		*(client->p++)= c;
		if(client->p-client->input>=80)
		{
		    client->state= 0;
		    client->p= client->input;
		}
	    }
	}
    }
}

static void broadcast(
const	uint8_t	*buffer,
	int	size
)
{
    Client *c= clients;
    while(c)
    {
	int n= size;
	Client *next= c->next;
	const uint8_t *p= buffer;
	while(n>0)
	{
	    int w= system::writeSocket(c->socket,p,n);
	    if(w>=0) n-= w;
	    else if(
		system::lastError()!=system::eAgain	&&
		system::lastError()!=system::eInterrupt
	    )
	    {
		removeClient(c);
		n= -1;
	    }
	}
	c= next;
    }
}

static inline uint32_t quickRnd(
uint32_t range
)
{
    static uint32_t state= 0x23653;
    state= (state*66049+3907864577U);
    return (uint32_t)((state/4294967295.0)*range);
}

static void shuffle(
int	nfd,
void	**sockets
)
{
    for(int i=0;i<nfd/2;++i)
    {
	int i1= quickRnd(nfd);
	int i2= quickRnd(nfd);
	void *s1= sockets[i1];
	void *s2= sockets[i2];
	sockets[i1]= s2;
	sockets[i2]= s1;
    }
}

static void usage()
{
    system::error(
	0,
	"Usage: gpsd [-readonly]                Prevent clients from writing to device\n"
        "            [-log <log file>]          Log all incoming connections to file\n"
        "            [-port <listen port>]      Which TCP port to listen for incoming\n"
        "            [-serial <serial port>]    Serial port of monitored device\n"
        "            [-speed <serial speed>]    Speed of serial port\n"
        "            [-help <recorded file>]    Dump help and exit\n"
        "            [-replay <recorded file>]  Replay  existing log\n\n"
    );
}

int main(
int	,
char	**argv
)
{
    const char *portA= 0;
    const char *speedA= 0;
    const char *serial= 0;
    const char *replay= 0;
    ++argv;

    system::error(0,"\nGPS3D V1.20  - gpsd - Please do not use for navigation.\n\n");

    while(1)
    {
	const char *arg= *(argv++);
	if(arg==0) break;

	if(system::strcmp(arg,"-port")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-port <listen port>\n");
	    portA= opt;
	}
	else if(system::strcmp(arg,"-speed")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-speed <serial speed>\n");
	    speedA= opt;
	}
	else if(system::strcmp(arg,"-serial")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-serial <serial port>\n");
	    serial= opt;
	}
	else if(system::strcmp(arg,"-readonly")==0)
	{
	    readOnly= 1;
	}
	else if(system::strcmp(arg,"-replay")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-replay <recorded file>\n");
	    replay= opt;
	    readOnly= 1;
	}
	else if(system::strcmp(arg,"-log")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-log <log file>\n");
	    logFile= opt;
	}
	else if(system::strcmp(arg,"-h")==0 || system::strcmp(arg,"-help"))
	{
	    usage();
	    system::exit(0);
	}
    }

    int port;
    int speed;
    if(serial==0) serial= system::defaultSerial();
    if(portA==0 || 1!=system::sscanf(portA,"%d",&port)) port= 2222;
    if(speedA==0 || 1!=system::sscanf(speedA,"%d",&speed)) speed= 19200;

    void *gps= replay==0 ? system::openSerial(serial,speed) : 0;
    if(gps==(void*)-1)	system::error(1,"fatal: Couldn't open serial port %s\n",serial);

    const char *mask= "0.0.0.0";
    void *lsock= system::makeListenSocket(mask,port);
    if(lsock==(void*)-1) system::error(1,"fatal: Couldn't create listening socket on port %s:%d\n",mask,port);

    system::daemonize();
    system::ignoreSignals();

    if(replay)
    {
	void **fds= 0;
	int fdsAlloc= 0;
	int replayFile= -1;
	while(1)
	{
	    int nfd= 1+clientCnt;
	    if(nfd>fdsAlloc)
	    {
		fdsAlloc= 2*nfd+10;
		fds= (void**)system::realloc(fds,fdsAlloc*sizeof(void*));
	    }

	    void **fp= fds;
	    *(fp++)= lsock;

	    Client *c= clients;
	    while(c)
	    {
		*(fp++)= c->socket;
		c= c->next;
	    }

	    shuffle(nfd,fds);

	    void *r= system::waitOn(nfd,fds,clients?1000:-1);
	    if(r==lsock)
	    {
		char buf[1024];
		void *asock= system::acceptConnection(lsock,buf);
		if(asock!=(void*)-1) addClient(asock,buf);
	    }
	    else if(r==(void*)-3)
	    {
		if(replayFile<0) replayFile= system::open(replay,system::readOnly);
		if(replayFile>=0)
		{
		    uint8_t buf[1024];
		    uint8_t *p= buf;
		    while(1)
		    {
			int r= system::read(replayFile,p,1);
			if(r!=1)
			{
			    system::close(replayFile);
			    replayFile= -1;
			    break;
			}
			else if(p[0]=='\n')
			{
			    broadcast((const uint8_t*)buf,p-buf+1);
			    break;
			}
			else
			{
			    ++p;
			}
		    }
		}
	    }
	    else
	    {
		c= clients;
		while(c)
		{
		    void *s= c->socket;
		    Client *next= c->next;
		    if(s==r)
		    {
			uint8_t buffer[2048];
			int r= system::readSocket(s,buffer,2048);
			if(r>0) clientInput(gps,c,buffer,r);
			else if(system::lastError()==system::eAgain) break;
			else if(system::lastError()!=system::eInterrupt)
			{
			    removeClient(c);
			    break;
			}
		    }
		    c= next;
		}
	    }
	}
    }

    void **fds= 0;
    int fdsAlloc= 0;
    while(1)
    {
	int nfd= 2+clientCnt;
	if(nfd>fdsAlloc)
	{
	    fdsAlloc= 2*nfd+10;
	    fds= (void**)system::realloc(fds,fdsAlloc*sizeof(void*));
	}

	void **fp= fds;
	*(fp++)= gps;
	*(fp++)= lsock;

	Client *c= clients;
	while(c)
	{
	    *(fp++)= c->socket;
	    c= c->next;
	}

	shuffle(nfd,fds);

	void *r= system::waitOn(nfd,fds);
	if(r==gps)
	{
	    while(1)
	    {
		uint8_t buffer[2048];
		int r= system::readSerial(gps,buffer,2048);
		if(r<=0) break;
		broadcast(buffer,r);
	    }
	}
	else if(r==lsock)
	{
	    char buf[1024];
	    void *asock= system::acceptConnection(lsock,buf);
	    if(asock!=(void*)-1) addClient(asock,buf);
	}
	else
	{
	    c= clients;
	    while(c)
	    {
		void *s= c->socket;
		Client *next= c->next;
		if(s==r)
		{
		    uint8_t buffer[2048];
		    int r= system::readSocket(s,buffer,2048);
		    if(r>0) clientInput(gps,c,buffer,r);
		    else if(system::lastError()==system::eAgain) break;
		    else if(system::lastError()!=system::eInterrupt)
		    {
			removeClient(c);
			break;
		    }
		}
		c= next;
	    }
	}
    }
    return 0;
}

