
#include <viz_app.h>
#include <viz_system.h>

#define GIF_CMAPMASK		0x80
#define GIF_EXTENSION		0x21
#define GIF_IMAGESEP		0x2c
#define GIF_INTERLACEMASK	0x40       
#define GIF_TRAILER		0x3b           

static const uint8_t ega[]=
{
    0x00,0x00,0x00, 0x80,0x00,0x00, 0x00,0x80,0x00, 0x80,0x80,0x00,
    0x00,0x00,0x80, 0x80,0x00,0x80, 0x00,0x80,0x80, 0xC8,0xC8,0xC8,
    0x64,0x64,0x64, 0xFF,0x64,0x64, 0x64,0xFF,0x64, 0xFF,0xFF,0x64,
    0x64,0x64,0xFF, 0xFF,0x64,0xFF, 0x64,0xFF,0xFF, 0xFF,0xFF,0xFF,
};

class Gif
{
public:
			Gif();
			~Gif();

	int		initLZW(int);
	int		streamLZW(int);
	void		outputPixel(uint8_t);
	int		stream(uint8_t *, int);

	int		w;
	int		h;
	int		flags;
	void		*socket;

	int		state;
	int		offset;
	int		blockSize;

	int		bitsLeft;
	uint32_t	bitBucket;

	int		tokenSize;
	int		eofToken;
	int		resetToken;
	int		maxToken;
	int		tokenMask;
	int		prevToken;
	int		firstChar;

	uint8_t		stack[4097];
	int		lzwDict[4096][2];
	int		initialTokenSize;

	int		x;
	int		y;
	int		rw;
	int		rh;
	int		pass;
	uint8_t		*rgb;
	uint8_t		*line;
	uint8_t		*colorMap;
	int		colorMapSize;

	int		cropBorder;
	double		la0;
	double		lo0;
	double		la1;
	double		lo1;
};

Gif::Gif()
{
    state=	0;
    socket=	(void*)-1;

    rgb=	0;
    colorMap=	0;
}

Gif::~Gif()
{
    if(rgb) system::free(rgb);
    if(colorMap) system::free(colorMap);
}

void Gif::outputPixel(
uint8_t index
)
{
    if(
	cropBorder<=y && y<h-cropBorder	&&
	cropBorder<=x && x<w-cropBorder
    )
    {
	int rx= x-cropBorder;
	uint8_t *dst= 3*rx + line;
	uint8_t *color= 3*index + colorMap;
	dst[0]= color[0];
	dst[1]= color[1];
	dst[2]= color[2];
    }

    ++x;
    if(x>=w)
    {
	if((flags&GIF_INTERLACEMASK)==0) ++y;
	else
	{
	    if(pass==0)
	    {
		y+= 8;
		if(y>=h)
		{
		    pass= 1;
		    y= 4;
		}
	    }
	    else if(pass==1)
	    {
		y+= 8;
		if(y>=h)
		{
		    pass= 2;
		    y= 2;
		}
	    }
	    else if(pass==2)
	    {
		y+= 4;
		if(y>=h)
		{
		    pass= 3;
		    y= 1;
		}
	    }
	    else if(pass==3)
	    {
		y+= 2;
	    }
	}
	line= 3*rw*(rh-1-(y-cropBorder)) + rgb;
	x= 0;
    }
}

int Gif::initLZW(
int c
)
{
    x= 0;
    y= 0;
    pass= 0;
    rw= w-2*cropBorder;
    rh= h-2*cropBorder;
    if(rw<0 || rh<0) return -1;
    rgb= (uint8_t*)system::malloc(3*rw*rh);
    line= 3*rw*(rh-1-(0-cropBorder)) + rgb;

    bitsLeft=		0;
    bitBucket=		0;

    tokenSize=		1+c;
    resetToken=		(1<<c);
    eofToken=		resetToken+1;
    maxToken=		resetToken+2;
    tokenMask=		(1<<tokenSize)-1;
    initialTokenSize=	tokenSize;
    prevToken=		-1;

    for(int i=0; i<resetToken; ++i)
    {
	lzwDict[i][0]= i;
	lzwDict[i][1]= 0;
    }
    return 0;
}

int Gif::streamLZW(
int c
)
{
    bitBucket|= (c<<bitsLeft);
    bitsLeft+= 8;

    uint8_t *top= stack;
    while(bitsLeft>=tokenSize)
    {
	int token= (bitBucket & tokenMask);
	bitBucket>>= tokenSize;
	bitsLeft-= tokenSize;

	if(token==eofToken) return -1;
	if(token>maxToken) return -1;

	if(token==resetToken)
	{
	    tokenSize= initialTokenSize;
	    tokenMask= (1<<tokenSize)-1;
	    maxToken= resetToken+2;
	    prevToken= -1;
	    continue;
	}

	if(prevToken==-1)
	{
	    outputPixel(lzwDict[token][0]);
	    prevToken= token;
	    firstChar= token;
	    continue;
	}

	int savedToken= token;
	if(token==maxToken)
	{
	    *(top++)= firstChar;
	    token= prevToken;
	}
	while(token>resetToken)
	{
	    *(top++)= lzwDict[token][0];
	    token= lzwDict[token][1];
	}

	*(top++)= firstChar= lzwDict[token][0];
	lzwDict[maxToken][0]= firstChar;
	lzwDict[maxToken][1]= prevToken;
	++maxToken;

	if((maxToken&tokenMask)==0 && maxToken<4096)
	{
	    tokenMask+= maxToken;
	    ++tokenSize;
	}
	prevToken= savedToken;

	while(1)
	{
	    --top;
	    outputPixel(top[0]);
	    if(top==stack) break;
	}
    }
    return 0;
}

int Gif::stream(
uint8_t *data,
int	n
)
{
    while(n--)
    {
	int c= *(data++);
	switch(state)
	{
	    case 0:			// <stuff>
		if(c=='\r')		state= 1;
		break;
	    case 1:			// <stuff>\r
		if(c=='\n')		state= 2;
		else if(c=='\r')	state= 1;
		else			state= 0;
		break;
	    case 2:			// <stuff>\r\n
		if(c=='\r')		state= 3;
		else			state= 0;
		break;
	    case 3:			// <stuff>\r\n\r
		if(c=='\n')		state= 4;
		else if(c=='\r')	state= 1;
		else			state= 0;
		break;
	    case 4:			// <stuff>\r\n\r\n
		if(c=='G')		state= 5;
		else			return -1;
		break;
	    case 5:			// <stuff>\r\n\r\nG
		if(c=='I')		state= 6;
		else			return -1;
		break;
	    case 6:			// <stuff>\r\n\r\nGI
		if(c=='F')		state= 7;
		else			return -1;
		break;
	    case 7:			// <stuff>\r\n\r\nGIF
		if(c!='8')		return -1;
		offset= 4;
		state= 8;
		break;
	    case 8:			// Reading header
		if(offset==6)		w= c;
		else if(offset== 7)	w|= (c<<8);
		else if(offset== 8)	h= c;
		else if(offset== 9)	h|= (c<<8);
		else if(offset==10)	flags= c;
		else if(offset==12)
		{
		    colorMapSize= (1<<(1+(flags&0x7)));
		    colorMap= (uint8_t*)system::malloc(3*colorMapSize);
		    if(flags&GIF_CMAPMASK)
		    {
			offset= -1;
			state= 9;
		    }
		    else
		    {
			uint8_t *d= colorMap;
			for(int i=0; i<colorMapSize; ++i)
			{
			    const uint8_t *s= ega + 3*(i&15);
			    d[0]= s[0];
			    d[1]= s[1];
			    d[2]= s[2];
			    d+=3;
			}
			state= 10;
		    }
		}
		++offset;
		break;
	    case 9:			// Reading global color map
		colorMap[offset++]= c;
		if(offset==3*colorMapSize) state= 10;
		break;
	    case 10:			// Reading block type
		switch(c)
		{
		    case GIF_EXTENSION:
			state= 11;
			break;
		    case GIF_IMAGESEP:
			offset= 0;
			state= 14;
			break;
		    default:
		    case GIF_TRAILER:
			return -1;
			break;
		}
		break;
	    case 11:			// Reading extension type
		state= 12;
		break;
	    case 12:			// Reading block size
		blockSize= c;
		if(blockSize==0)	state= 10;
		else			state= 13;
		break;
	    case 13:			// Skipping block data
		--blockSize;
		if(blockSize==0) state= 12;
		break;
	    case 14:			// Just got an IMAGE_SEP, reading raster header
		if(offset==4)		w= c;
		else if(offset==5)	w|= (c<<8);
		else if(offset==6)	h= c;
		else if(offset==7)	h|= (c<<8);
		else if(offset==8)
		{
		    flags= c;
		    if((flags&GIF_CMAPMASK)==0) state= 16;
		    else
		    {
			colorMapSize= (1<<(1+(flags&0x7)));
			colorMap= (uint8_t*)system::realloc(colorMap,3*colorMapSize);
			offset= -1;
			state= 15;
		    }
		}
		++offset;
		break;
	    case 15:			// Reading local colormap
		colorMap[offset++]= c;
		if(offset==3*colorMapSize) state= 16;
		break;
	    case 16:			// Colormap done, init LZW
		if(initLZW(c))		return -1;
		state=  17;
		break;
	    case 17:			// Read data block length
		blockSize= c;
		if(blockSize==0)	return -1;
		else			state= 18;
		break;
	    case 18:
		--blockSize;
		if(streamLZW(c)<0) return -1;
		if(blockSize==0) state= 17;
		break;
	}
    }
    return 0;
}

int gifHandler(
void *data
)
{
    Gif *gif= (Gif*)data;
    while(1)
    {
	uint8_t buffer[16384];
	int r= system::readSocket(
	    gif->socket,
	    buffer,
	    16384
	);

	if(r>0 && gif->stream(buffer,r)<0) r= -1;
	else if(r<0 && system::lastError()==system::eAgain)	break;
	else if(r<0 && system::lastError()==system::eInterrupt)	continue;

	if(r<=0)
	{
	    makePatch(
		gif->la0,
		gif->lo0,
		gif->la1,
		gif->lo1,
		gif->rgb,
		gif->rw,
		gif->rh
	    );
	    delete gif;
	    return -1;
	}
    }
    return 0;
}

void *newGif(
void	*socket,
double	la0,
double	lo0,
double	la1,
double	lo1,
int	cb
)
{
    Gif *gif= new Gif();
    gif->socket= socket;
    gif->cropBorder= cb;
    gif->la0= la0;
    gif->lo0= lo0;
    gif->la1= la1;
    gif->lo1= lo1;
    return gif;
}

