
#include <viz_app.h>
#include <viz_system.h>

static int checksum(
const uint8_t *s
)
{
    uint8_t sum1= 0;
    while(1)
    {
	uint8_t c= *(s++);
	if(c==0 || c=='*') break;
	sum1^= c;
    }

    uint32_t sum2= 0;
    while(1)
    {
	int c= *(s++);
	c-= (c>'9' ? 'A'-10 : '0');
	if(c<0 || c>15) break;
	sum2<<= 4;
	sum2+= c;
    }

    return sum2!=(uint32_t)sum1;
}

static void sentence(
const uint8_t *s
)
{
    if(checksum(s) && getFixSums()==0) return;

    int nFields= 0;
    const uint8_t *src= s;
    uint8_t fields[81][81];
    uint8_t *dst= fields[1];
    while(1)
    {
	int c= *(src++);
	if(c!=0 && c!=',' && c!='*') *(dst++)= c;
	else
	{
	    dst[0]= 0;
	    if(c==0) break;
	    else dst= fields[1+(nFields++)];
	}
    }

    if(
	s[0]=='G'	&&
	s[1]=='P'	&&
	s[2]=='G'	&&
	s[3]=='G'	&&
	s[4]=='A'
    )
    {
	newGGA(fields[1],fields[6],fields[9]);
    }
    else if(
	s[0]=='G'	&&
	s[1]=='P'	&&
	s[2]=='G'	&&
	s[3]=='L'	&&
	s[4]=='L'
    )
    {
	if(fields[6][0]=='A') newFix(fields[1],fields[2],fields[3],fields[4],fields[5]);
    }
    else if(
	s[0]=='G'	&&
	s[1]=='P'	&&
	s[2]=='G'	&&
	s[3]=='S'	&&
	s[4]=='V'
    )
    {
	if(nFields> 4 && fields[ 4][0])	newSatelliteFix(fields[ 4],fields[ 5],fields[ 6],fields[ 7]);
	if(nFields> 8 && fields[ 8][0])	newSatelliteFix(fields[ 8],fields[ 9],fields[10],fields[11]);
	if(nFields>12 && fields[12][0]) newSatelliteFix(fields[12],fields[13],fields[14],fields[15]);
	if(nFields>16 && fields[16][0])	newSatelliteFix(fields[16],fields[17],fields[18],fields[19]);
    }
#if 0
    else if(
	s[0]=='P'	&&
	s[1]=='M'	&&
	s[2]=='G'	&&
	s[3]=='N'	&&
	s[4]=='T'	&&
	s[5]=='R'	&&
	s[6]=='K'
    )
    {
	//if(fields[8][0]=='A') newTrackFix(fields[1],fields[2],fields[3],fields[4]);
	newTrackFix(fields[1],fields[2],fields[3],fields[4]);
    }
#endif
    else if(
	s[0]=='P'	&&
	s[1]=='M'	&&
	s[2]=='G'	&&
	s[3]=='N'
#if 0
	s[4]=='W'	&&
	s[5]=='P'	&&
	s[6]=='L'
#endif
    )
    {
	newWaypoint(fields[1],fields[2],fields[3],fields[4],fields[5],fields[6],fields[7],fields[8],fields[9]);
    }
    else if(
	s[0]=='G'	&&
	s[1]=='P'	&&
	s[2]=='G'	&&
	s[3]=='S'	&&
	s[4]=='A'
    )
    {
	newSolution(fields,nFields);
    }
}

void gpsCharInput(
uint8_t c
)
{
    static uint8_t *p;
    static int state= 0;
    static uint8_t buffer[81];
    switch(state)
    {
	case 0:
	    if(c=='\n')		state= 1;
	    break;
	case 1:
	    if(c=='$')		state= 2;
	    else if(c=='\n')	state= 1;
	    else		state= 0;
	    p= buffer;
	    break;
	case 2:
	    if(c=='\r')
	    {
		p[0]= 0;
		state= 0;
		sentence(buffer);
	    }
	    else
	    {
		*(p++)= c;
		if(p-buffer>=80) state= 0;
	    }
	    break;
    }
}

int gpsInput()
{
    while(1)
    {
	uint8_t c;
	int r= system::readSocket(gpsSocket(),&c,1);
	if(r==1) gpsCharInput(c);
	else break;
    }
    return 0;
}

void *gpsSocket()
{
    static int init= 0;
    static void *socket;
    if(init==0)
    {
	int port;
	const char *sport= getPort();
	const char *server= getServer();
	if(1!=system::sscanf(sport,"%d",&port)) port= 2222;
	socket= system::makeSocket(server,port);
	if(socket==(void*)-1)	system::error(0,"warning: can't connect to GPS server %s:%d\n",server,port);
	else			system::error(0,"info: Connected to GPS server %s:%d\n",server,port);
	init= 1;
    }
    return socket;
}

void gpsOutput(
const char *msg
)
{
    uint8_t buf[160];
    system::sprintf((char*)buf,"$PMGNCMD,%s*XX\r\n",msg);

    uint8_t sum= 0;
    uint8_t *p= buf+1;
    while(1)
    {
	uint8_t c= *(p++);
	if(c=='*') break;
	sum^= c;
    }

    static const uint8_t *tohex= (const uint8_t *)"0123456789ABCDEF";
    p[0]= tohex[(sum>>4)&0xF];
    p[1]= tohex[(sum>>0)&0xF];

    int n= 4+p-buf;
    while(n)
    {
	int r= system::writeSocket(gpsSocket(),buf,n);
	if(r>=0) n-= r;
	else if(system::lastError()!=system::eAgain && system::lastError()!=system::eInterrupt)
	{
	    system::error(0,"warning: can't write to gps socket\n");
	    break;
	}
    }
}

void gpsLoadTrackFile(
const char *fileName
)
{
    int ok= 0;
    int sz= system::fileSize(fileName);
    if(sz>0)
    {
	int track= system::open(fileName,system::readOnly);
	if(track>=0)
	{
	    uint8_t *buf= (uint8_t*)system::malloc(sz+1);
	    if(sz==system::read(track,buf,sz))
	    {
		uint8_t *p0= 0;
		uint8_t *p= buf;
		uint8_t *end= buf+sz;
		while(1)
		{
		    if(p[0]=='$') p0= p+1;
		    else if(p[0]=='\n' || p==end)
		    {
			p[0]= 0;
			sentence(p0);
			if(p==end) break;
		    }
		    ++p;
		}
		ok= 1;
	    }
	    system::close(track);
	    system::free(buf);
	}
    }
    if(ok==0) system::error(0,"Couldn't load track file (%s)\n", fileName);
}

