
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

#define MAXZOOM 500
static double zoomLevel= 50;

static double zoomFactor(
double x
)
{
    double m= x/30.0;
    return 1*0.7 + m*(2000*0.7-1*0.7);
}

static double zoomScale(
double x
)
{
    double m= x/30.0;
    return (int)(1000 + m*(2000000-1000));
}

void zoomHTTPMap(
int up
)
{
    if(up)			zoomLevel*= 1.25;
    else if(zoomLevel>1e-10)	zoomLevel*= 0.80;

    if(zoomLevel>MAXZOOM)	zoomLevel= MAXZOOM;
    else if(zoomLevel<1e-5)	zoomLevel= 1e-5;

    trash();
}

void getMapSquare(
double	*la0,
double	*lo0,
double	*la1,
double	*lo1
)
{
    double la= lastPickLat();
    double lo= lastPickLon();
    double loPixelScale= zoomFactor(zoomLevel)/15000000.0;	// Radian per pixel
    double laPixelScale= zoomFactor(zoomLevel)/15000000.0;	// Radian per pixel
    loPixelScale*= 180.0/system::pi()/system::cos(la*system::pi()/180.0);
    laPixelScale*= 180.0/system::pi();

    int width= 256;
    int height= 256;
    la0[0]= la-laPixelScale*height/2.0;
    lo0[0]= lo-loPixelScale*width/2.0;
    la1[0]= la+laPixelScale*height/2.0;
    lo1[0]= lo+loPixelScale*width/2.0;
}

void paintMapSquare()
{
    if(getMapMode())
    {

	double la0;
	double lo0;
	double la1;
	double lo1;
	getMapSquare(&la0,&lo0,&la1,&lo1);

	glColor3ub(0x00,0x00,0xFF);
	paintSquare(la0,lo0,la1,lo1);

	double laMid=(la0+la1)/2.0;
	double loMid=(lo0+lo1)/2.0;
	paintGreatcircle(la0,lo0,laMid,loMid,0,0,0);
	paintGreatcircle(laMid,loMid,la1,lo1,0,0,0);
	paintGreatcircle(la1,lo0,laMid,loMid,0,0,0);
	paintGreatcircle(laMid,loMid,la0,lo1,0,0,0);
    }
}

static double findGoodZoomLevel(
double	eyeDist
)
{
    double z= eyeDist*800.0;
    if(z>1500.0) z= 1500.0;
    return z;
}

void getHTTPMap(
int	useDefault,
double	forceLa,
double	forceLo,
double	dist
)
{
    int cb= 20;
    int iconType= 10;
    int width= 256+2*cb;
    int height= 256+2*cb;
    double la= useDefault ? lastPickLat() : forceLa;
    double lo= useDefault ? lastPickLon() : forceLo;

    double iconLa= la;
    double iconLo= lo;
    double zl= useDefault ? zoomLevel : findGoodZoomLevel(dist);
    double zoom= zoomFactor(zl);
    double scale= zoomScale(zl);
    double loPixelScale= zoom/15000000.0;	// Radian per pixel
    double laPixelScale= zoom/15000000.0;	// Radian per pixel

    width-= 2*cb;
    height-= 2*cb;
    loPixelScale*= 180.0/system::pi()/system::cos(la*system::pi()/180.0);
    laPixelScale*= 180.0/system::pi();

    double la0= la-laPixelScale*height/2.0;
    double lo0= lo-loPixelScale*width/2.0;
    double la1= la+laPixelScale*height/2.0;
    double lo1= lo+loPixelScale*width/2.0;

    char url[1024];
    system::sprintf(	
	url,
	"/gif?&CT=%.8g:%.8g:%.8g&IC=%.6g:%.6g:%d:&W=%d&H=%d&FAM=myblast&LB=",
	la,
	lo,
	scale,
	iconLa,
	iconLo,
	iconType,
	width,
	height
    );

system::error(0,"http://www.mapblast.com%s\n",url);

    void *socket= system::makeSocket("www.mapblast.com",80);
    if(socket==(void*)-1) return;

    char buffer[2048];
    system::sprintf(
	buffer,
	"GET %s HTTP/1.0\r\n"
	"Connection: Keep-Alive\r\n"
	"User-Agent: GPS3d (http://www.mgix.com/gps3d)\r\n"
	"Accept: image/gif\r\n"
	"\r\n",
	url
    );

    char *p= buffer;
    int n= system::strlen(buffer);
    while(n>0)
    {
	int r= system::writeSocket(
	    socket,
	    p,
	    n
	);
	if(r>0)								{ n-= r; p+= r; }
	else if(r==0 && system::lastError()==system::eNone)		continue;
	else if(r< 0 && system::lastError()==system::eAgain)		continue;
	else if(r< 0 && system::lastError()==system::eInterrupt)	continue;
	else
	{
	    system::closeSocket(socket);
	    return;
	}
    }

    void *gif= newGif(socket,la0,lo0,la1,lo1,cb);
    newChannel(socket,gifHandler,gif);
}

