
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

static int init= 0;

void loadMap()
{
    if(init)	return;

    glBindTexture(GL_TEXTURE_2D,1);
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);

    const char *fileName= MAP_FILE;    
    int sz= system::fileSize(fileName);
    if(sz<=0) return;

    int f= system::open(fileName,system::readOnly);
    if(f<0) return;

    uint32_t w;
    uint32_t h;
    uint8_t *map0;
    uint8_t *map= map0= (uint8_t*)system::malloc(sz);
    if(sz==system::read(f,map,sz))
    {
	if(map[0]=='P' && map[1]=='6' && (map[2]=='\r' || map[2]=='\n'))
	{
	    if(2==system::sscanf((const char*)(map+(map[2]=='\r'?3:2)),"%d %d",&w,&h))
	    {
		int i=0;
		while(1)
		{
		    int c= *(map++);
		    if(map-map0>sz) break;
		    if(c=='\n' && ++i==3) break;
		}
		if(i==3)
		{
		    int pixelSize= w*h;
		    system::error(0,"info: loaded map %s (%dx%d)\n",fileName,w,h);
		    if(sz-(map-map0)==3*pixelSize)	init= 1;
		    else				system::error(0,"warning:map file %s has incorrect size\n",fileName);
		}
	    }
	}
    }
    system::close(f);

    if(init==0)
    {
	init= 1;
	system::error(0,"warning:error loading map %s\n",fileName);
	return;
    }

    GLint max;
    uint32_t mw= w;
    uint32_t mh= h;
    uint32_t level= 0;
    uint8_t *buffer= (uint8_t*)system::malloc(w*h);
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max);
    while(mw>0 && mh>0)
    {
	if(mw<(uint32_t)max && mh<(uint32_t)max)
	{
	    uint8_t *mip;
	    if(mw==w && mh==h)	mip= map;
	    else		resample(mip= buffer,mw,mh,map,w,h);
	    glTexImage2D(	
		GL_TEXTURE_2D,
		level++,
		3,
		mw,
		mh,
		0,
		GL_RGB,
		GL_UNSIGNED_BYTE,
		mip
	    );
	}
	mw>>= 1;
	mh>>= 1;
    }
    system::free(buffer);
    system::free(map0);
}

