
#include <viz_app.h>
#include <viz_system.h>

void apply44To3(
	double	*o,
const   double	*m,
const   double	*p 
)
{
    double cx;
    double cy;
    double cz;
    double px=p[0];
    double py=p[1];
    double pz=p[2];

    cx = m[ 0]*px;
    cx+= m[ 1]*py;
    cx+= m[ 2]*pz;
    cx+= m[ 3];

    cy = m[ 4]*px;
    cy+= m[ 5]*py;
    cy+= m[ 6]*pz;
    cy+= m[ 7];

    cz = m[ 8]*px;
    cz+= m[ 9]*py;
    cz+= m[10]*pz;
    cz+= m[11];

    o[0]= cx;
    o[1]= cy;
    o[2]= cz;
}

void apply44To3v(
	double	*o,
const   double	*m,
const   double	*p 
)
{
    double cx;
    double cy;
    double cz;
    double px=p[0];
    double py=p[1];
    double pz=p[2];

    cx = m[ 0]*px;
    cx+= m[ 1]*py;
    cx+= m[ 2]*pz;

    cy = m[ 4]*px;
    cy+= m[ 5]*py;
    cy+= m[ 6]*pz;

    cz = m[ 8]*px;
    cz+= m[ 9]*py;
    cz+= m[10]*pz;

    o[0]= cx;
    o[1]= cy;
    o[2]= cz;
}

void trsp44(
double *matrix
)
{
    double i1= matrix[ 1];
    double i2= matrix[ 2];
    double i3= matrix[ 3];

    double i4= matrix[ 4];
    double i6= matrix[ 6];
    double i7= matrix[ 7];

    double i8= matrix[ 8];
    double i9= matrix[ 9];
    double iB= matrix[11];

    double iC= matrix[12];
    double iD= matrix[13];
    double iE= matrix[14];

    matrix[ 1]= i4;
    matrix[ 2]= i8;
    matrix[ 3]= iC;
    matrix[ 4]= i1;
    matrix[ 6]= i9;
    matrix[ 7]= iD;
    matrix[ 8]= i2;
    matrix[ 9]= i6;
    matrix[11]= iE;
    matrix[12]= i3;
    matrix[13]= i7;
    matrix[14]= iB;
}

void rot44(
	double  *matrix,
const	double  *origin,
const	double  *vector,
	double  angle
)
{
    double x=vector[0];
    double y=vector[1];
    double z=vector[2];

    double o1=origin[0];
    double o2=origin[1];
    double o3=origin[2];

    double a= angle*system::pi()/180.0;
    double c= system::cos(a);
    double s= system::sin(a);
    double t= 1-c;

    double xy= x*y;
    double xz= x*z;
    double yz= y*z;

    double sx= s*x;
    double sy= s*y;
    double sz= s*z;

    matrix[ 0]= t*x*x + c;
    matrix[ 1]= t*xy - sz;
    matrix[ 2]= t*xz + sy;
    matrix[ 3]= o1 - o1*matrix[0] - o2*matrix[1] - o3*matrix[2];

    matrix[ 4]= t*xy + sz;
    matrix[ 5]= t*y*y + c;
    matrix[ 6]= t*yz - sx;
    matrix[ 7]= o2 - o1*matrix[4] - o2*matrix[5] - o3*matrix[6];

    matrix[ 8]= t*xz - sy;
    matrix[ 9]= t*yz + sx;
    matrix[10]= t*z*z + c;
    matrix[11]= o3 - o1*matrix[8] - o2*matrix[9] - o3*matrix[10];

    matrix[12]= 0.0;
    matrix[13]= 0.0;
    matrix[14]= 0.0;
    matrix[15]= 1.0;
}

