
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

static int mode;
static int state;
static int firstX;
static int firstY;
static int help=1;
static int width= 720;
static int height= 576;
static int filtering= 1;
static int mapMode= 1;

int windowHeight()		{ return height;	}
int windowWidth()		{ return width;		}
int getHelpState()		{ return help;		}
int getDisplayMode()		{ return mode;		}
int getInteractionState()	{ return state;		}
int getFiltering()		{ return filtering;	}
int getMapMode()		{ return mapMode;	}

void onWindowSize(
int w,
int h
)
{
    glViewport(0,0,width=w,height=h);
    trash();
}

void onWindowDamage()
{
    trash();
}

void onWindowClose()
{
    system::exit(0);
}

void onButtonDown(
int	,
int	x,
int	y,
int	shift,
int	ctrl,
int	alt
)
{
    if(state==0)
    {
	saveView();
	firstX= x;
	firstY= y;
	state=
	    1				|
	    (shift	? 2 : 0)	|
	    (ctrl	? 4 : 0)	|
	    (alt	? 8 : 0);
	if(state==3) startArea(x,y);
	trash();
    }
}

void onMouseMove(
int	x,
int	y
)
{
    if(state==0)	// No button down --> Picking
    {
	pick(x,y);
    }
    else if(state==1)	// No modifiers	--> Tumble
    {
	double dx= 540.0*(firstX-x)/(double)width;
	double dy= 540.0*(firstY-y)/(double)height;
	tumble(dy,dx);
    }
    else if(state==3)	// Shift --> Select
    {
	moveArea(x,y);
    }
    else if(state==5)	// Ctrl	--> Zoom
    {
	double f= 1.0+2.0*(x-firstX)/(double)width;
	if(f<1.0f) f=1/(2.0-f);
	zoom(f);
    }
    else if(state==7)	// Ctrl+Shift --> Pan
    {
	double dx= (x-firstX)/(double)width;
	double dy= (y-firstY)/(double)height;
	pan(dx,dy);
    }
}

void onButtonUp(
int	,
int	x,
int	y
)
{
    if(state==3) endArea(x,y);
    state= 0;
    trash();
}

void onKeyDown(
int	key,
int	,
int	,
int	,
int	,
int
)
{
    switch(key)
    {
	case 'h':
	case 'H':
	    help^=1;
	    trash();
	    break;
	case 't':
	case 'T':
	    gpsOutput("TRACK");
	    break;
	case 'w':
	case 'W':
	    gpsOutput("WAYPOINT");
	    break;
	case 'l':
	case 'L':
	    gpsLoadTrackFile(TRACK_FILE);
	    break;
	case 'c':
	case 'C':
	    centerViewOnFix(lastLat(),lastLon());
	    break;
	case 'p':
	case 'P':
	    centerViewOnFix(lastPickLat(),lastPickLon());
	    break;
	case 'm':
	case 'M':
	    mapMode^= 1;
	    trash();
	    break;
	case system::EnterKey:
	    getHTTPMap();
	    break;
	case 'g':
	case 'G':
	    getHTTPBrushData();
	    break;
	case 'x':
	case 'X':
	    zoomHTTPMap(1);
	    break;
	case 'z':
	case 'Z':
	    zoomHTTPMap(0);
	    break;
	case 'f':
	case 'F':
	    filtering^= 1;
	    trash();
	    break;
	case 'q':
	case 'Q':
	case system::EscapeKey:
	    system::exit(0);
	case 'r':
	case 'R':
	case system::HomeKey:
	    resetView();
	    break;
	case 'A':
	case 'a':
	    toggleAutomap();
	    break;
	case '+':
	case '=':
	    saveView();
	    zoom(0.8);
	    break;
	case '-':
	case '_':
	    saveView();
	    zoom(1.25);
	    break;
	case system::TabKey:
	    mode^= 1;
	    trash();
	    break;
	case ',':
	    setSubdivision(getSubdivision()-1);
	    break;
	case '.':
	    setSubdivision(getSubdivision()+1);
	    break;
    }
}

