
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

class Patch
{
public:
    int		txNum;
    double	la0;
    double	lo0;
    double	la1;
    double	lo1;
    double	pos0[3];
    double	pos1[3];
    double	pos2[3];
    double	pos3[3];
};

#define MAXPATCHES 8
static Patch *patches[MAXPATCHES];

static int txCounter=1;
static int nPatches;

void makePatch(
double		la0,
double		lo0,
double		la1,
double		lo1,
uint8_t		*rgb,
uint32_t	w,
uint32_t	h
)
{
    if(nPatches>=MAXPATCHES)
    {
	GLuint tid= (GLuint)patches[0]->txNum;;
	glDeleteTextures(1,&tid);
	delete patches[0];
	--nPatches;

	for(int i=0;i<nPatches;++i) patches[i]= patches[i+1];
    }

    Patch *patch= patches[nPatches++]= new Patch;
    patch->la0= la0;
    patch->lo0= lo0;
    patch->la1= la1;
    patch->lo1= lo1;

    localFrame(la0,lo0,patch->pos0,0,0);
    localFrame(la0,lo1,patch->pos1,0,0);
    localFrame(la1,lo1,patch->pos2,0,0);
    localFrame(la1,lo0,patch->pos3,0,0);

    glBindTexture(GL_TEXTURE_2D,patch->txNum=++txCounter);
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);

    int bits= 0;
    int maxDim= w>h ? w : h;
    while(maxDim)
    {
	maxDim>>= 1;
	++bits;
    }
    maxDim= 1<<(bits);

    int level= 0;
    uint8_t *map= (uint8_t*)system::malloc(3*maxDim*maxDim);
    while(maxDim>0)
    {
	resample(map,maxDim,maxDim,rgb,w,h);
	glTexImage2D(
	    GL_TEXTURE_2D,
	    level++,
	    3,
	    maxDim,
	    maxDim,
	    0,
	    GL_RGB,
	    GL_UNSIGNED_BYTE,
	    map
	);
	maxDim>>= 1;
    }
    system::free(map);
    trash();
}

void paintPatches()
{
    for(int i=0; i<nPatches; ++i)
    {
	glCullFace(GL_BACK);
	glColor3ub(0xFF,0xFF,0xFF);

	Patch *p= patches[i];
	glBindTexture(GL_TEXTURE_2D,p->txNum);

	double loDelta= system::fabs(p->lo1-p->lo0);
	double laDelta= system::fabs(p->la1-p->la0);
	if(loDelta<0.5 && laDelta<0.5)
	{
	    glBegin(GL_QUADS);
		glTexCoord2d(0.0,0.0);
		glVertex3dv(p->pos0);

		glTexCoord2d(1.0,0.0);
		glVertex3dv(p->pos1);

		glTexCoord2d(1.0,1.0);
		glVertex3dv(p->pos2);

		glTexCoord2d(0.0,1.0);
		glVertex3dv(p->pos3);
	    glEnd();
	}
	else
	{
	    for(int iy=0; iy<10; ++iy)
	    {
		double y= iy/10.0;
		double la1= p->la0 + (y+0.0)*(p->la1-p->la0);
		double la2= p->la0 + (y+0.1)*(p->la1-p->la0);
		for(int ix=0; ix<10; ++ix)
		{
		    double x= ix/10.0;
		    double lo1= p->lo0 + (x+0.0)*(p->lo1-p->lo0);
		    double lo2= p->lo0 + (x+0.1)*(p->lo1-p->lo0);

		    double r0[3];
		    double r1[3];
		    double r2[3];
		    double r3[3];

		    localFrame(la1,lo1,r0,0,0);
		    localFrame(la1,lo2,r1,0,0);
		    localFrame(la2,lo2,r2,0,0);
		    localFrame(la2,lo1,r3,0,0);

		    glBegin(GL_QUADS);
			glTexCoord2d(x+0.0,y+0.0);
			glVertex3dv(r0);

			glTexCoord2d(x+0.1,y+0.0);
			glVertex3dv(r1);

			glTexCoord2d(x+0.1,y+0.1);
			glVertex3dv(r2);

			glTexCoord2d(x+0.0,y+0.1);
			glVertex3dv(r3);
		    glEnd();
		}
	    }
	}
	glColor3ub(0x00,0x00,0xFF);
	paintSquare(p->la0,p->lo0,p->la1,p->lo1);
    }
}

