
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

#define TRAJSIZE 10000
static int trajSizes[50];
static int trajColor[50];
static double lastSatAzimuth[50];
static double lastSatElevation[50];
static double satX[50][TRAJSIZE];
static double satY[50][TRAJSIZE];
static double satZ[50][TRAJSIZE];
static double satT[50][TRAJSIZE];

// earthRadius = 6 378 136 meters
// GPS satellites orbit at 20 180 000 meters altitude (from ground)= 26558136 from earth center
static double altitudeSquared= 17.33834459756036121946;

int sphereIntersect(
double	*r1,
double	*r2,
double	radiusSquared,
double	*r,
double	*d
)
{
    int valid= 0;
    double a= d[0]*d[0] + d[1]*d[1] + d[2]*d[2];
    double b= 2.0*(r[0]*d[0] + r[1]*d[1] + r[2]*d[2]);
    double c= r[0]*r[0] + r[1]*r[1] + r[2]*r[2] - radiusSquared;

    double delta= b*b-4*a*c;
    if(delta>=0)
    {
	double lambda1= (-b+system::sqrt(delta))/(2.0*a);
	double lambda2= (-b-system::sqrt(delta))/(2.0*a);
	if(r1)
	{
	    r1[0]= r[0] + lambda1*d[0];
	    r1[1]= r[1] + lambda1*d[1];
	    r1[2]= r[2] + lambda1*d[2];
	}
	if(r2)
	{
	    r2[0]= r[0] + lambda2*d[0];
	    r2[1]= r[1] + lambda2*d[1];
	    r2[2]= r[2] + lambda2*d[2];
	}
	++valid;
    }
    return valid;
}

void newSolution(
const uint8_t	sol[][81],
int		nFields
)
{
    int tj[50];
    int size= 50*sizeof(int);
    system::memset(tj,0,size);
    if(nFields>15) nFields=15;
    for(int j=3; j<nFields; ++j)
    {
	if(sol[j][0])
	{
	    int sat;
	    if(1==system::sscanf((const char*)sol[j],"%d",&sat)) tj[sat]=1;
	}
    }

    if(system::memcmp(trajColor,tj,size))
    {
	system::memcpy(trajColor,tj,size);
	trash();
    }
}

static int ae2xyz(
double	lat,
double	lon,
double	azimuth,
double	elevation,
double	aSquared,
double	*result
)
{
    double r[3];			// Normed radial vector (==pos on sphere, since radius==1)
    double u[3];			// Normed tangent vector to latitude towards the east
    double v[3];			// Normed tangent vector to meridians towards the north
    localFrame(lat,lon,r,u,v);

    azimuth*= system::pi()/180.0;
    elevation*= system::pi()/180.0;

    double vc= system::cos(elevation)*system::cos(azimuth);
    double uc= system::cos(elevation)*system::sin(azimuth);
    double rc= system::sin(elevation);

    double d[3];			// Direction from origin to satellite
    d[0]= uc*u[0] + vc*v[0] + rc*r[0];
    d[1]= uc*u[1] + vc*v[1] + rc*r[1];
    d[2]= uc*u[2] + vc*v[2] + rc*r[2];

    return sphereIntersect(result,0,aSquared,r,d);
}

void newSatelliteFix(
const uint8_t	*prn,
const uint8_t	*elevation,
const uint8_t	*azimuth,
const uint8_t	*
)
{
    int num;	
    double e;
    double a;
    if(lastStamp()<0) return;
    if(1!=system::sscanf((const char*)prn,"%d",&num)) return;
    if(1!=system::sscanf((const char*)elevation,"%lf",&e)) return;
    if(1!=system::sscanf((const char*)azimuth,"%lf",&a)) return;

    if(e>90.0) return;
    if(a>360.0) return;
    if(num>=50) return;

    int pnum= trajSizes[num];
    if(
	pnum==0				||
	lastSatAzimuth[num]!=a		||
	lastSatElevation[num]!=e
    )
    {
	double sat[3];
	double la= lastLat();
	double lo= lastLon();
	int correct= ae2xyz(
	    la,
	    lo,
	    a,
	    e,
	    altitudeSquared,
	    sat
	);
	if(correct)
	{
	    int mod= pnum%TRAJSIZE;
	    satX[num][mod]= sat[0];
	    satY[num][mod]= sat[1];
	    satZ[num][mod]= sat[2];
	    satT[num][mod]= lastStamp();
	    trajSizes[num]= ++pnum;
	    trash();
	}
    }
}

static inline void satPos(
int	i,
int	j,
double	*p
)
{
    p[0]= satX[i][j];
    p[1]= satY[i][j];
    p[2]= satZ[i][j];
}

void paintSatellites()
{
    for(int i=0;i<50;++i)
    {
	int start= 0;
	int nPoints= trajSizes[i];
	if(nPoints>TRAJSIZE)
	{
	    start= nPoints%TRAJSIZE;
	    nPoints= 10000;
	}

	if(trajColor[i])	glColor3ub(0x0,0xFF,0x0);
	else			glColor3ub(0xFF,0x00,0x0);

	double p[3];
	if(nPoints==1)
	{
	    glBegin(GL_POINTS);
		satPos(i,0,p);
		glVertex3dv(p);
	    glEnd();
	}
	else if(nPoints)
	{
	    glBegin(GL_LINE_STRIP);
		for(int j=0; j<nPoints; ++j)
		{
		    int jx= ((j+start)%TRAJSIZE);
		    satPos(i,jx,p);
		    glVertex3dv(p);
		}
	    glEnd();

	    uint8_t buf[100];
	    system::sprintf((char*)buf,"%02d",i);
	    glRasterPos3dv(p);
	    paintString(buf);
	}
    }
}

