
#include <viz_app.h>
#include <viz_gl.h>
#include <viz_system.h>

#define CZ 0.866025403784438646763723170752936183471402626905190314027903	// cos(30)
#define SZ 0.500000000000000000000000000000000000000000000000000000000000	// sin(30)
#define C1 0.951056516295153572116439333379382143405698634125750222447305	// cos(18)
#define S1 0.309016994374947424102293417182819058860154589902881431067723	// sin(18)
#define C2 0.587785252292473129168705954639072768597652437643145991072272	// cos(54)
#define S2 0.809016994374947424102293417182819058860154589902881431067724	// sin(54)

#define X1 (C1*CZ)
#define Y1 (S1*CZ)
#define X2 (C2*CZ)
#define Y2 (S2*CZ)

#define ico_p0	0.0,	1.0,	0.0
#define ico_p1	-X2,	SZ,	-Y2
#define ico_p2	X2,	SZ,	-Y2
#define ico_p3	X1,	SZ,	Y1
#define ico_p4	0,	SZ,	CZ
#define ico_p5	-X1,	SZ,	Y1

#define ico_m0	-X1,	-SZ,	-Y1
#define ico_m1	0,	-SZ,	-CZ
#define ico_m2	X1,	-SZ,	-Y1
#define ico_m3	X2,	-SZ,	Y2
#define ico_m4	-X2,	-SZ,	Y2
#define ico_m5	0.0,	-1.0,	0.0

static double icosaedron[]=
{
    ico_p1, ico_p0, ico_p2,
    ico_p5, ico_p0, ico_p1,
    ico_p4, ico_p0, ico_p5,
    ico_p3, ico_p0, ico_p4,
    ico_p2, ico_p0, ico_p3,

    ico_p1, ico_m1, ico_m0,
    ico_m0, ico_p5, ico_p1,
    ico_p5, ico_m0, ico_m4,
    ico_m4, ico_p4, ico_p5,
    ico_p4, ico_m4, ico_m3,
    ico_m3, ico_p3, ico_p4,
    ico_p3, ico_m3, ico_m2,
    ico_m2, ico_p2, ico_p3,
    ico_p2, ico_m2, ico_m1,
    ico_m1, ico_p1, ico_p2,

    ico_m3, ico_m5, ico_m2,
    ico_m4, ico_m5, ico_m3,
    ico_m0, ico_m5, ico_m4,
    ico_m1, ico_m5, ico_m0,
    ico_m2, ico_m5, ico_m1,
};

static void vertex(
double	*ip,
double	**op,
double *last
)
{
    double x= ip[0];
    double y= ip[1];
    double z= ip[2];
    double n= system::sqrt(x*x+y*y+z*z);
    x/= n;
    y/= n;
    z/= n;

    double phi;
    double theta;
    if(y==1.0)
    {
	phi= 1.0;
	theta= last[0] - 0.1;	// 36 degrees
    }
    else if(y==-1.0)
    {
	phi= 0.0;
	theta= last[0] + 0.1;	// 36 degrees
    }
    else
    {
	phi= 0.5 + system::asin(y)/system::pi();
	theta= 0.5 + system::atan2(x,z)/(2.0*system::pi());
    }

    if(last[0]<=1000.0)
    {
	double d= theta-last[0];
	if(d>1.0/6.0)		theta-= 1.0;
	else if(d<-1.0/6.0)	theta+= 1.0;
    }
    last[0]= theta;

    (*op)[0]= x;
    (*op)[1]= y;
    (*op)[2]= z;
    (*op)[3]= theta;
    (*op)[4]= phi;
    (*op)+= 5;
}

static inline void lerp(
double	*v,
double	*a,
double	*b,
double	mu
)
{
    v[0]= a[0] + mu*(b[0]-a[0]);
    v[1]= a[1] + mu*(b[1]-a[1]);
    v[2]= a[2] + mu*(b[2]-a[2]);
}

void sphere(
int	subd,
int	*_nStrips,
int	**_stripSizes,
double	**_strips
)
{
    int nStrips= 20*subd;
    int nVertices= 20*(subd*3 + subd*(subd-1));
    int *stripSizes= (int*)system::malloc(nStrips*sizeof(int));
    double *strips= (double*)system::malloc(nVertices*5*sizeof(double));

    double *p= strips;
    int *s= stripSizes;
    double *icos= icosaedron;
    for(int i=0; i<20; ++i)
    {
	double *p0= icos + 0;
	double *p1= icos + 3;
	double *p2= icos + 6;
	for(int j=0; j<subd; ++j)
	{
	    double mu;
	    double v0[3];
	    double v1[3];
	    double v2[3];
	    double v3[3];
	    double va[3];
	    double vb[3];

	    mu= (j+1)/(double)subd;
	    lerp(v0,p1,p0,mu);
	    lerp(v2,p1,p2,mu);

	    mu= j/(double)subd;
	    lerp(v1,p1,p0,mu);
	    lerp(v3,p1,p2,mu);

	    double last= 4000.0;
	    vertex(v0,&p,&last);
	    vertex(v1,&p,&last);
	    for(int k=0; k<j; ++k)
	    {
		lerp(va,v0,v2,(k+1)/(double)(j+1));
		lerp(vb,v1,v3,(k+1)/(double)j);
		vertex(va,&p,&last);
		vertex(vb,&p,&last);
	    }
	    vertex(v2,&p,&last);
	    *(s++)= 3+2*j;
	}
	icos+= 9;
    }

    _nStrips[0]= nStrips;
    _stripSizes[0]= stripSizes;
    _strips[0]= strips;
}

static int nStrips;
static int *stripSizes;
static double *strips;
static int subdivision= 20;

static void trashSphere()
{
    if(strips) system::free(strips);
    if(stripSizes) system::free(stripSizes);

    strips= 0;
    stripSizes= 0;

    trash();
}

int getSubdivision()
{
    return subdivision;
}

void setSubdivision(
int s
)
{
    subdivision= (s>=1?s:1);
    trashSphere();
}

void paintSphere()
{
    if(strips==0) sphere(subdivision,&nStrips,&stripSizes,&strips);

    glCullFace(GL_BACK);
    glColor3ub(0xFF,0xFF,0xFF);
    glBindTexture(GL_TEXTURE_2D,1);

    double *p= strips;
    for(int s= 0; s<nStrips; ++s)
    {
	int size= stripSizes[s];
	glBegin(GL_TRIANGLE_STRIP);
	while(size--)
	{
	    glTexCoord2dv(p+3);
	    glNormal3dv(p);
	    glVertex3dv(p);
	    p+= 5;
	}
	glEnd();
    }
}

