#define __SYSTEM_EXPOSE_GUTS__
#include <viz_system.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

const int system::readOnly=	O_RDONLY|O_BINARY;
const int system::writeOnly=	O_WRONLY|O_BINARY;
const int system::creatFlag=	O_CREAT;
const int system::truncFlag=	O_TRUNC;

int system::sprintf(
	char *buf,
const	char *format,
...
)
{
    int ret;
    va_list arg;
    va_start(arg,format);
	ret= vsprintf(buf,format,arg);
    va_end(arg);
    return ret;
}

void system::error(
	int	level,
const	char	*format,
...
)
{
    va_list arg;
    va_start(arg,format);
	vfprintf(stdout,format,arg);
    va_end(arg);
    if(level==1) exit(1);
}

int system::fileSize(
const	char	*name
)
{
    struct stat buf;
    int r=stat(name,&buf);
    if(r>=0) r= buf.st_size;
    return r;
}


#ifndef WIN32
int system::sscanf(
const	char *buf,
const	char *format,
...
)
{
    int ret;
    va_list arg;
    va_start(arg,format);
	ret= vsscanf(buf,format,arg);
    va_end(arg);
    return ret;
}
#else
/*
 * The things you gotta do on NT ... pain,pain,pain
 */
static void *truesscanf= (void*)sscanf;

__declspec(naked) int system::sscanf(
const char *,
const char *,
...
)
{
    _asm
    {
	jmp	truesscanf
    }
}
#endif // WIN32

#ifdef WIN32
int WINAPI WinMain(
HINSTANCE   ,
HINSTANCE   ,
LPSTR	    cmd,
int
)
{
    extern int main(int,const char**);
    const char *argv[3];
    argv[0]= "gpsd.exe";
    argv[1]= cmd;
    argv[2]= 0;
    return main(cmd && cmd[0] ? 2 : 1,argv);
}

class IOChannel
{
public:
			IOChannel()	{}
    virtual		~IOChannel()	{}
    virtual HANDLE	getWaitable()=	0;
protected:
};

class SerialChannel:public IOChannel
{
public:
			    SerialChannel(HANDLE s);
    virtual		    ~SerialChannel();
    virtual HANDLE	    getWaitable();
    virtual int		    read(void*,size_t);
    virtual int		    write(const void*,size_t);
	    DWORD	    readerThread();

private:
	    HANDLE	serial;
	    HANDLE	thread;
	    HANDLE	waitable;

	    size_t	size;
	    uint8_t	*buffer;
	    uint8_t	*pointer;
	    uint32_t	spinLock;
};

SerialChannel::SerialChannel(
HANDLE s
)
{
    serial= s;
    thread= 0;
    waitable= 0;

    size= 0;
    buffer= 0;
    pointer= 0;
    spinLock= 0;
}

static uint8_t asmLockCode[]=
{
    0x31,	// xorl   %eax,%eax		eax=0
    0xc0,
    0x8b,	// movl   0x4(%esp,1),%ebx	ebx=lock addr
    0x5c,
    0x24,
    0x04,
    0xb9,	// movl   $0x1,%ecx		ecx=1
    0x01,
    0x00,
    0x00,
    0x00,
    0xf0,	// lock cmpxchgl %ecx,(%ebx)
    0x0f,
    0xb1,
    0x0b,
    0x75,	// jne lock
    0xfa,
    0xc3	// ret
};

static void (*asmLock)(uint32_t*)= (void (*)(uint32_t*))(void*)asmLockCode;

int SerialChannel::read(
void	*p,
size_t	n
)
{
    (*asmLock)(&spinLock);
	size_t available= pointer-buffer;
	size_t obtained= n>available ? available : n;
	if(obtained)
	{
	    available-= obtained;
	    memcpy(p,buffer,obtained);
	    memmove(buffer,buffer+obtained,available);
	    pointer= buffer+available;
	}
    spinLock= 0;
    return obtained;
}

int SerialChannel::write(
const	void	*p,
	size_t	n
)
{
    DWORD cnt;
    BOOL r= WriteFile(serial,p,n,&cnt,0);
    return r==FALSE ? -1 : cnt;
}

DWORD SerialChannel::readerThread()
{
    while(1)
    {
	DWORD cnt;
	uint8_t b;
	if(ReadFile(serial,&b,1,&cnt,0) && cnt==1)
	{
	    (*asmLock)(&spinLock);
		size_t offset= pointer-buffer;
		if(offset>=size)
		{
		    size= 2*(size+10);
		    buffer= (uint8_t*)system::realloc(buffer,size);
		    pointer= buffer+offset;
		}

		*(pointer++)= b;
		SetEvent(waitable);
	    spinLock= 0;
	}
    }
    return 0;
}

static DWORD __stdcall serialHandler(
void *s
)
{
    return ((SerialChannel*)s)->readerThread();
}

HANDLE SerialChannel::getWaitable()
{
    if(thread==0)
    {
	DWORD id= 0;
	waitable= CreateEvent(0,0,0,0);
	thread= CreateThread(0,4096,serialHandler,this,0,&id);
    }
    return waitable;
}

SerialChannel::~SerialChannel()
{
    CloseHandle(serial);
    if(thread)
    {
	TerminateThread(thread,0);
	CloseHandle(thread);
	CloseHandle(waitable);
    }
    if(buffer) system::free(buffer);
}

class SocketChannel:public IOChannel
{
public:
		    SocketChannel(SOCKET s);
    virtual	    ~SocketChannel();
    virtual HANDLE  getWaitable();
    virtual int	    read(void*,size_t);
    virtual int	    write(const void*,size_t);
private:
    SOCKET	socket;
    HANDLE	event;
};

SocketChannel::SocketChannel(
SOCKET	s
)
{
    socket= s;
    event= 0;
}

int SocketChannel::read(
void	*p,
size_t	n
)
{
    int r= recv(socket,(char*)p,n,0);
    return r==SOCKET_ERROR ? -1 : r;
}

int SocketChannel::write(
const	void	*p,
	size_t	n
)
{
    int r= send(socket,(const char*)p,n,0);
    return r==SOCKET_ERROR ? -1 : r;
}

HANDLE SocketChannel::getWaitable()
{
    if(event==0) event= CreateEvent(0,0,0,0);
    WSAEventSelect(socket,event,FD_READ);
    return event;
}

SocketChannel::~SocketChannel()
{
    CloseHandle(event);
    closesocket(socket);
}

class ListenChannel:public IOChannel
{
public:
		    ListenChannel(SOCKET s);
    virtual	    ~ListenChannel();
    virtual HANDLE  getWaitable();
	    void    *accept(char *);
private:
    SOCKET	socket;
    HANDLE	event;
};

ListenChannel::ListenChannel(
SOCKET	s
)
{
    socket= s;
    event= 0;
}

HANDLE ListenChannel::getWaitable()
{
    if(event==0) event= CreateEvent(0,0,0,0);
    WSAEventSelect(socket,event,FD_ACCEPT);
    return event;
}

void *ListenChannel::accept(
char *buf
)
{
    SOCKADDR addr;
    int length= sizeof(addr);
    SOCKET s= ::accept(socket,&addr,&length);

    struct sockaddr_in *sai= (struct sockaddr_in *)&addr;
    sprintf(buf,"%s",inet_ntoa(sai->sin_addr));
    return s==SOCKET_ERROR ? (void*)-1 : (void*) new SocketChannel(s);
}

ListenChannel::~ListenChannel()
{
    CloseHandle(event);
    closesocket(socket);
}

static void initWinsock()
{
    static int initDone= 0;
    if(initDone==0)
    {
	WSADATA data;
	WORD version= MAKEWORD(2,2);
	int err=WSAStartup(version,&data);
	if(err || data.wVersion!=version) system::error(0,"warning:can't init winsock dll\n");
	initDone= 1;
    }
}
#endif // WIN32

void *system::makeSocket(
const	char	    *server,
	uint32_t    port
)
{
#ifdef WIN32
    initWinsock();

    struct hostent *servHE= gethostbyname(server);
    if(servHE==0) return (void*)-1;

    SOCKET s= socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if(s==INVALID_SOCKET) return (void*)-1;

    BOOL o;
    LINGER lng;
    system::memset(&lng,0,sizeof(lng));
    if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(const char*)&o,sizeof(o))==SOCKET_ERROR) return (void*)-1;
    if(setsockopt(s,SOL_SOCKET,SO_LINGER,(const char*)&lng,sizeof(lng))==SOCKET_ERROR) return (void*)-1;

    SOCKADDR_IN addr;
    memset(&addr,0,sizeof(addr));
    memcpy(&(addr.sin_addr),servHE->h_addr,servHE->h_length);
    addr.sin_family= servHE->h_addrtype;
    addr.sin_port= htons(port);

    if(connect(s,(const sockaddr*)&addr,sizeof(addr))==SOCKET_ERROR) return (void*)-1;
    return (void*)new SocketChannel(s);
#else
    struct hostent *servHE= gethostbyname(server);
    if(servHE==0) return (void*)-1;

    int s= socket(AF_INET,SOCK_STREAM,0);
    if(s<0) return (void*)-1;

    int o= 1;
    struct linger lng;
    system::memset(&lng,0,sizeof(lng));
    if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&o,sizeof(o))<0) return (void*)-1;
    if(setsockopt(s,SOL_SOCKET,SO_LINGER,&lng,sizeof(lng))<0) return (void*)-1;

    struct sockaddr_in addr;
    memset(&addr,0,sizeof(addr));
    memcpy(&(addr.sin_addr),servHE->h_addr,servHE->h_length);
    addr.sin_family= servHE->h_addrtype;
    addr.sin_port= htons((short)port);

    if(connect(s,(struct sockaddr*)&addr,sizeof(addr))<0) return (void*)-1;
    if(fcntl(s,F_SETFL,O_NONBLOCK)<0) return (void*)-1;
    return (void*)s;
#endif
}

void *system::makeListenSocket(
const	char	    *netmask,
	uint32_t    port
)
{
#ifdef WIN32
    initWinsock();

    SOCKADDR_IN lport;
    memset(&lport,0,sizeof(lport));
    lport.sin_family=		AF_INET;
    lport.sin_port=		htons((uint16_t)port);
    lport.sin_addr.S_un.S_addr=	htonl(inet_addr(netmask));

    SOCKET s= socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if(s==INVALID_SOCKET) return (void*)-1;
    if(bind(s,(struct sockaddr*)&lport,sizeof(lport))==SOCKET_ERROR) return (void*)-1;
    if(listen(s,100)==SOCKET_ERROR) return (void*)-1;

    BOOL o;
    LINGER lng;
    system::memset(&lng,0,sizeof(lng));
    if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(const char*)&o,sizeof(o))==SOCKET_ERROR) return (void*)-1;
    if(setsockopt(s,SOL_SOCKET,SO_LINGER,(const char*)&lng,sizeof(lng))==SOCKET_ERROR) return (void*)-1;
    return (void*)new ListenChannel(s);
#else
    struct sockaddr_in lport;
    memset(&lport,0,sizeof(lport));
    lport.sin_family=		AF_INET;
    lport.sin_port=		htons((uint16_t)port);
    lport.sin_addr.s_addr=	htonl(inet_addr(netmask));

    int s= socket(AF_INET,SOCK_STREAM,0);
    if(s<0) return (void*)-1;
    if(bind(s,(struct sockaddr*)&lport,sizeof(lport))<0) return (void*)-1;
    if(listen(s,100)) return (void*)-1;

    int o= 0;
    struct linger lng;
    memset(&lng,0,sizeof(lng));
    if(setsockopt(s,SOL_SOCKET,SO_LINGER,&lng,sizeof(lng))<0) return (void*)-1;
    if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&o,sizeof(o))<0) return (void*)-1;
    return (void*)s;
#endif
}

int system::readSocket(
void	*s,
void	*p,
size_t	n
)
{
#ifdef WIN32
    if(s==(void*)-1) return -1;
    return ((SocketChannel*)s)->read(p,n);
#else
    return ::read((long)s,p,n);
#endif
}

int system::writeSocket(
	void	*s,
const	void	*p,
	size_t	n
)
{
#ifdef WIN32
    if(s==(void*)-1) return -1;
    return ((SocketChannel*)s)->write(p,n);
#else
    return ::write((long)s,p,n);
#endif
}

void *system::acceptConnection(
void *lsock,
char *buf
)
{
#ifdef WIN32
    if(lsock==(void*)-1) return (void*)-1;
    return ((ListenChannel*)lsock)->accept(buf);
#else
    struct sockaddr_in addr;
    socklen_t len= sizeof(struct sockaddr_in);
    void *retVal= (void*)accept((long)lsock,(struct sockaddr*)&addr,&len);
    sprintf(buf,"%s",inet_ntoa(addr.sin_addr));
    return retVal;
#endif
}

void system::closeSocket(
void	*s
)
{
#ifdef WIN32
    if(s==(void*)-1) return;
    else if(s==(void*)-2) return;
    else delete (IOChannel*)s;
#else
    close((long)s);
#endif
}

void *system::openSerial(
const	char	*name,
	int	speed
)
{
#ifdef WIN32
    HANDLE fd= CreateFile(
	name,
	GENERIC_READ|GENERIC_WRITE,
	0,
	0,
	OPEN_EXISTING,
	0,
	0
    );
    if(fd==INVALID_HANDLE_VALUE) return (void*)-1;

    DCB dcb;
    GetCommState(fd, &dcb);
    dcb.BaudRate= speed;
    dcb.ByteSize= 8;
    dcb.Parity= NOPARITY;
    dcb.StopBits= ONESTOPBIT;
    dcb.fOutxCtsFlow= FALSE;
    dcb.fOutxDsrFlow= FALSE;
    dcb.fDtrControl= DTR_CONTROL_DISABLE;
    dcb.fTXContinueOnXoff= TRUE;
    dcb.fOutX= FALSE;
    dcb.fInX= FALSE;
    dcb.fNull= 0;
    dcb.fRtsControl= RTS_CONTROL_DISABLE;
    dcb.fAbortOnError= FALSE;
    if(SetCommState(fd,&dcb)==FALSE) system::error(0,"warning:couldn't configure serial port %s\n",name);
    return (void*)new SerialChannel(fd);
#else
    const char *tty= name ? name : "/dev/ttyS0";
    int fd= open(tty,O_NOCTTY|O_RDWR|O_NDELAY);
    if(fd>0)
    {
	int msp= 0;
	switch(speed)
	{
	    case 75:	    msp= B75;	    break;
	    case 150:	    msp= B150;	    break;
	    case 300:	    msp= B300;	    break;
	    case 600:	    msp= B600;	    break;
	    case 1200:	    msp= B1200;	    break;
	    case 2400:	    msp= B2400;	    break;
	    case 4800:	    msp= B4800;	    break;
	    case 9600:	    msp= B9600;	    break;
	    case 19200:	    msp= B19200;    break;
	    case 38400:	    msp= B38400;    break;
	    case 57600:	    msp= B57600;    break;
	    case 115200:    msp= B57600;    break;
	}
	struct termios st;
	st.c_iflag= IGNBRK|IGNPAR;
	st.c_oflag= 0;
	st.c_cflag= CS8|CREAD|CLOCAL;
	st.c_lflag= 0;
	cfsetispeed(&st,msp);
	cfsetospeed(&st,msp);
	if(tcsetattr(fd,TCSANOW,&st)<0) system::error(0,"warning:couldn't configure serial port %s\n",name);
    }
    return (void*)fd;
#endif
}

int system::readSerial(
void	*s,
void	*p,
size_t	n
)
{
#ifdef WIN32
    if(s==(void*)-1) return -1;
    return ((SerialChannel*)s)->read(p,n);
#else
    return ::read((long)s,p,n);
#endif
}

int system::writeSerial(
	void	*s,
const	void	*p,
	size_t	n
)
{
#ifdef WIN32
    if(s==(void*)-1) return -1;
    return ((SerialChannel*)s)->write(p,n);
#else
    return ::write((long)s,p,n);
#endif
}

void system::closeSerial(
void *s
)
{
#ifdef WIN32
    if(s==(void*)-1) return;
    else if(s==(void*)-2) return;
    else delete (IOChannel*)s;
#else
    close((long)s);
#endif
}

const char *system::defaultSerial()
{
#ifdef WIN32
    return "COM1";
#else
    return "/dev/ttyS0";
#endif
}

void *system::waitOn(
int	n,
void	**channels,
int	timeOut
)
{
#ifdef WIN32
    int i;
    int cnt= 0;
    int msg= 0;
    HANDLE *events=(HANDLE*)alloca(n*sizeof(HANDLE));
    HANDLE *handles=(HANDLE*)alloca(n*sizeof(HANDLE));
    for(i=0;i<n;++i)
    {
	void *s= channels[i];
	if(s==(void*)-2) msg= 1;
	else if(s!=(void*)-1)
	{
	    IOChannel *ioc= (IOChannel*)channels[i];
	    events[cnt]= ioc->getWaitable();
	    handles[cnt]= channels[i];
	    ++cnt;
	}
    }

    if(msg)
    {
	if(cnt==0)
	{
	    return (void*)-2;
	    DWORD r= MsgWaitForMultipleObjects(
		0,
		0,
		FALSE,
		timeOut<0 ? INFINITE : timeOut,
		QS_ALLINPUT
	    );
	    if(r==WAIT_TIMEOUT)	return (void*)-3;
	    else if(r==WAIT_OBJECT_0+cnt) return (void*)-2;
	    else return (void*)-1;
	}
	else
	{
	    DWORD r= MsgWaitForMultipleObjects(
		cnt,
		events,
		FALSE,
		timeOut<0 ? INFINITE : timeOut,
		QS_ALLINPUT
	    );
	    if(r==WAIT_TIMEOUT)	return (void*)-3;
	    else if(r==WAIT_OBJECT_0+cnt) return (void*)-2;
	    else if(WAIT_OBJECT_0<=r && r<WAIT_OBJECT_0+cnt) return (void*)handles[r-WAIT_OBJECT_0];
	}
    }
    else if(cnt>0)
    {
	DWORD r= WaitForMultipleObjectsEx(
	    cnt,
	    events,
	    FALSE,
	    timeOut<0 ? INFINITE : timeOut,
	    FALSE
	);
	if(r==WAIT_TIMEOUT) return (void*)-3;
	else if(WAIT_OBJECT_0<=r && r<WAIT_OBJECT_0+cnt) return (void*)handles[r-WAIT_OBJECT_0];
    }
    return (void*)-1;
#else
    int i;
    int max= -1;
    fd_set readSet;
    FD_ZERO(&readSet);
    for(i=0;i<n;++i)
    {
	long s= (long)channels[i];
	if(s>=0)
	{
	    FD_SET(s,&readSet);
	    if(max<s) max= s;
	}
    }

    struct timeval tout;
    tout.tv_sec= timeOut/1000;
    tout.tv_usec= (timeOut%1000)*1000;

    int r= select(1+max,&readSet,0,0,timeOut<0?0:&tout);
    if(r==0) return (void*)-3;
    else if(r<0) system::error(0,"warning: select failed\n");

    for(i=0;i<n;++i)
    {
	long s= (long)channels[i];
	if(s>=0 && FD_ISSET(s,&readSet)) return (void*)s;
    }
    return (void*)-1;
#endif
}

void system::daemonize()
{
#ifdef WIN32
    // FIXME: do it
#else
    int pid= fork();
    if(pid==-1) system::error(1,"Can't fork");
    else if(pid) exit(0);
    setpgid(0,0);
    close(0);
    close(1);
    close(2);
#endif
}

void system::ignoreSignals()
{
#ifdef WIN32
#else
    signal(SIGPIPE,SIG_IGN);
    signal(SIGALRM,SIG_IGN);
    signal(SIGCHLD,SIG_IGN);
#endif
}

int system::lastError()
{
#ifdef WIN32
    int err= GetLastError();
    if(err==0)		    return system::eNone;
    if(err==WSAEWOULDBLOCK) return system::eAgain;
    if(err==WSAEINPROGRESS) return system::eAgain;
    if(err==WSAEINTR)	    return system::eInterrupt;
    return system::eOther ;
#else
    int err= errno;
    if(err==0)		    return system::eNone;
    if(err==EAGAIN)	    return system::eAgain;
    if(err==EINTR)	    return system::eInterrupt;
    return system::eOther ;
#endif
}
