#ifndef __viz_system__H
#define __viz_system__H

#include <viz_types.h>

#ifdef _DEBUG
#    define INLINE
#    define ASSERT(x) { if((x)==0){ system::error("fatal:assertion failed file %s line %d\n",__FILE__,__LINE__); ((int*)1)[0]= 0;} }
#else
#    define INLINE inline
#    define ASSERT(x)
#    ifndef __SYSTEM_EXPOSE_GUTS__
#        define __SYSTEM_EXPOSE_GUTS__
#    endif
#endif

class system
{
public:
    INLINE  static double	pi();
    INLINE  static double	fabs(double);
    INLINE  static double	sqrt(double);
    INLINE  static double	cos(double);
    INLINE  static double	sin(double);
    INLINE  static double	tan(double);
    INLINE  static double	acos(double);
    INLINE  static double	asin(double);
    INLINE  static double	atan2(double,double);

    INLINE  static int		chdir(const char*);
    INLINE  static int		open(const char*,int,int =0);
    INLINE  static int		read(int,void*,size_t);
    INLINE  static int		write(int,const void*,size_t);
    INLINE  static void		close(int);

    INLINE  static void		exit(int);
    INLINE  static void		*malloc(size_t);
    INLINE  static void		*realloc(void*,size_t);
    INLINE  static void		free(void*);
    INLINE  static char		*strdup(const char *);
    INLINE  static void		memset(void *,int,size_t);
    INLINE  static int		memcmp(void *,void*,size_t);
    INLINE  static void		memcpy(void *,void*,size_t);
    INLINE  static int		strcmp(const char*,const char*);
    INLINE  static int		strlen(const char*);

	    static int		lastError();
	    static int		fileSize(const char*);
	    static void		error(int,const char*,...);
	    static int		sprintf(char*,const char*,...);
	    static int		sscanf(const char*,const char*,...);

	    static void		*makeSocket(const char*,uint32_t);
	    static void		*makeListenSocket(const char*,uint32_t);
	    static void		*acceptConnection(void*,char* =0);
	    static int		readSocket(void*,void*,size_t);
	    static int		writeSocket(void*,const void*,size_t);
	    static void		closeSocket(void*);

	    static void		*waitOn(int,void **,int =-1);
	    static void		ignoreSignals();
	    static void		daemonize();

	    static void		*openSerial(const char*,int);
	    static int		readSerial(void*,void*,size_t);
	    static int		writeSerial(void*,const void*,size_t);
	    static void		closeSerial(void*);
	    static const char	*defaultSerial();

    static  const   int		readOnly;
    static  const   int		writeOnly;
    static  const   int		creatFlag;
    static  const   int		truncFlag;

	enum
	{
	    EscapeKey= 0x100,
	    HomeKey,
	    TabKey,
	    EnterKey
	};

	enum
	{
	    eNone= 0,
	    eAgain,
	    eInterrupt,
	    eOther,
	};
};

#ifdef __SYSTEM_EXPOSE_GUTS__

#ifdef linux
extern "C" {
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
extern int vsscanf(const char *,const char *,va_list); // Here until glibc's headers get it right
}
#endif // linux

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <fcntl.h>
#include <io.h>
#include <direct.h>
#include <malloc.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <winsock2.h >
#endif // WIN32

#ifdef macosx
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <termios.h>
typedef int socklen_t;
#endif // macosx

INLINE	double	system::pi()				    { return 3.141592653589793238462643383279502884197169399375105820;  }
INLINE	double	system::fabs(double x)			    { return ::fabs(x);							}
INLINE	double	system::sqrt(double x)			    { return ::sqrt(x);							}
INLINE	double	system::cos(double x)			    { return ::cos(x);							}
INLINE	double	system::sin(double x)			    { return ::sin(x);							}
INLINE	double	system::tan(double x)			    { return ::tan(x);							}
INLINE	double	system::acos(double x)			    { return ::acos(x);							}
INLINE	double	system::asin(double x)			    { return ::asin(x);							}
INLINE	double	system::atan2(double y,double x)	    { return ::atan2(y,x);						}

INLINE	int	system::chdir(const char *n)	    	    { return ::chdir(n);						}
INLINE	int	system::open(const char *n,int f,int m)	    { return ::open(n,f,m);						}
INLINE	int	system::read(int f,void *p,size_t n)	    { return ::read(f,p,n);						}
INLINE	int	system::write(int f,const void *p,size_t n) { return ::write(f,p,n);						}
INLINE	void	system::close(int f)			    { ::close(f);							}

INLINE	void	system::exit(int n)			    { ::exit(n);							}
INLINE	void	*system::malloc(size_t sz)		    { return ::malloc(sz);						}
INLINE	void	*system::realloc(void *p,size_t sz)	    { return ::realloc(p,sz);						}
INLINE	void	system::free(void *p)			    { ::free(p);							}
INLINE	void	system::memset(void *p,int c,size_t n)	    { ::memset(p,c,n);							}
INLINE	int	system::memcmp(void *a,void *b,size_t n)    { return ::memcmp(a,b,n);						}
INLINE	void	system::memcpy(void *d,void *s,size_t n)    { ::memcpy(d,s,n);							}
INLINE	char	*system::strdup(const char *s)		    { return ::strdup(s);						}
INLINE	int	system::strcmp(const char *a,const char *b) { return ::strcmp(a,b);						}
INLINE	int	system::strlen(const char *s)		    { return ::strlen(s);						}

#endif // __SYSTEM_EXPOSE_GUTS__
#endif // __viz_system_H__
