
#include <viz_app.h>
#include <viz_system.h>

#ifdef WIN32

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL/gl.h>

static HDC	dc;
static HGLRC	glc;
static HWND	window;

static void initMess(
HWND hwnd
)
{
    PIXELFORMATDESCRIPTOR pfd;
    pfd.nSize=		sizeof(PIXELFORMATDESCRIPTOR );
    pfd.nVersion=	1;
    pfd.cDepthBits=	32;
    pfd.iLayerType=	PFD_MAIN_PLANE;
    pfd.dwFlags=	PFD_DRAW_TO_WINDOW		|
			PFD_SUPPORT_OPENGL		|
			PFD_DOUBLEBUFFER;
    pfd.iPixelType=	PFD_TYPE_RGBA;
    pfd.cColorBits=	32;
    pfd.cRedBits=	8;
    pfd.cGreenBits=	8;
    pfd.cBlueBits=	8;
    pfd.cDepthBits=	32;

    dc= GetDC(hwnd);
    if(dc==NULL) system::error(1,"fatal: can't get window DC\n");

    int fmt= ChoosePixelFormat(dc,&pfd);
    if(fmt==0) system::error(1,"fatal: can't get a decent OpenGL pixel format\n");

    BOOL r= SetPixelFormat(dc,fmt,&pfd);
    if(r==FALSE) system::error(1,"fatal: can't set window pixel format\n");

    glc= wglCreateContext(dc);
    if(glc==NULL) system::error(1,"fatal: can't create OpenGL context\n");
 
    BringWindowToTop(hwnd);
    EnableWindow(hwnd,TRUE);
    SetStretchBltMode(dc,COLORONCOLOR);
    wglMakeCurrent(dc,glc);
}

static LONG WINAPI winProc( 
HWND	hwnd, 
UINT	type, 
WPARAM	wparam, 
LPARAM	lparam
) 
{
    PAINTSTRUCT p;
    switch(type)
    {
	case WM_CREATE:
	    initMess(hwnd);
	    onWindowSize(
		((CREATESTRUCT*)lparam)->x,
		((CREATESTRUCT*)lparam)->y
	    );
	    break;
	case WM_SIZE:
	    onWindowSize(
		(int)LOWORD(lparam),
		(int)HIWORD(lparam)
	    );
	    break;
	case WM_PAINT:
	    BeginPaint(hwnd,&p);
	    EndPaint(hwnd,&p);
	    onWindowDamage();
	    break;
	case WM_CLOSE:
	    onWindowClose();
	    break;
	case WM_ERASEBKGND:
	    break;
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	    SetFocus(hwnd);
	    SetActiveWindow(hwnd);
	    SetForegroundWindow(hwnd);
	    SetCapture(hwnd);
	    onButtonDown(
		type==WM_LBUTTONDOWN	?	0	:
		type==WM_MBUTTONDOWN	?	1	:
		type==WM_RBUTTONDOWN	?	2	:
		-1,
		LOWORD(lparam),
		HIWORD(lparam),
		(GetKeyState(VK_SHIFT)<0),
		(GetKeyState(VK_CONTROL)<0),
		(GetKeyState(VK_MENU)<0)
	    );
	    break;
	case WM_MOUSEMOVE:
	    onMouseMove(
		(int)LOWORD(lparam),
		(int)HIWORD(lparam)
	    );
	    break;
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	    ReleaseCapture();
	    onButtonUp(
		type==WM_LBUTTONDOWN	?	0	:
		type==WM_MBUTTONDOWN	?	1	:
		type==WM_RBUTTONDOWN	?	2	:
		-1,
		LOWORD(lparam),
		HIWORD(lparam)
	    );
	    break;
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
	    if(wparam!=VK_HOME) goto def;
	    onKeyDown(system::HomeKey,0,0,0,0,0);
	    break;
	case WM_CHAR:
	case WM_SYSCHAR:
	    if(wparam==27)	    wparam=system::EscapeKey;
	    else if(wparam=='\t')   wparam=system::TabKey;
	    else if(wparam=='\n')   wparam=system::EnterKey;
	    else if(wparam==13)	    wparam=system::EnterKey;
	    onKeyDown(wparam,0,0,0,0,0);
	    break;
	default:
	def:
	    return DefWindowProc(hwnd,type,wparam,lparam);
    }
    return 1;
}

void *eventSocket()
{
    static int init= 0;
    if(init==0)
    {
	WNDCLASSEX wc;
	ZeroMemory(&wc,sizeof(wc));
	wc.cbSize=		sizeof(wc);
	wc.style=		CS_DBLCLKS|CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc=		winProc;
	wc.hInstance=		GetModuleHandle(NULL);
	wc.hIcon=
	wc.hIconSm=		LoadIcon(GetModuleHandle(NULL),"Example");
	wc.hCursor=		LoadCursor(NULL,IDC_ARROW);
	wc.lpszClassName=	"VIZ";
	RegisterClassEx(&wc);

	window= CreateWindowEx(
	    WS_EX_ACCEPTFILES,
	    "VIZ",
	    "VIZ",
	    WS_VISIBLE|WS_OVERLAPPEDWINDOW,
	    100, 100, 720, 576,
	    0,
	    NULL,
	    NULL,
	    0
	);
	init= 1;
    }
    return (void*)-2;	// bogus socket that system::waitFor will handle
}

int windowInput()
{
    MSG msg;
    while(PeekMessage(&msg,NULL,0,0,PM_NOREMOVE))
    {
	GetMessage(&msg,NULL,0,0);
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }
    return 0;
}

void flush()
{
    glFlush();
    SwapBuffers(dc);
}

#endif // WIN32

#ifdef X11

#include <GL/glx.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

static Window window;
static Display *display;

static int convertX11Key(
int key
)
{
    switch(key)
    {
	case XK_Tab:	key= system::TabKey;	break;
	case XK_Home:	key= system::HomeKey;	break;
	case XK_Escape:	key= system::EscapeKey;	break;
	case XK_Return:	key= system::EnterKey;	break;
    }
    return key;
}

int windowInput()
{
    while(1)
    {
	XEvent event;
	XNextEvent(display, &event);
	switch(event.type)
	{
	    case ConfigureNotify:
		onWindowSize(
		    event.xconfigure.width,
		    event.xconfigure.height
		);
		break;
	    case Expose:
		onWindowDamage();
		break;
	    case ButtonPress:
		onButtonDown(
		    event.xbutton.button-1,
		    event.xbutton.x,
		    event.xbutton.y,
		    (event.xbutton.state&ShiftMask) ? 1 : 0,
		    (event.xbutton.state&ControlMask) ? 1 : 0,
		    (event.xbutton.state&Mod1Mask) ? 1 : 0
		);
		break;
	    case MotionNotify:
		onMouseMove(
		    event.xmotion.x,
		    event.xmotion.y
		);
		break;
	    case ButtonRelease:
		onButtonUp(
		    event.xbutton.button-1,
		    event.xbutton.x,
		    event.xbutton.y
		);
		break;
	    case KeyPress:
		int key= XLookupKeysym(&(event.xkey), 0);
		key= convertX11Key(key);
		onKeyDown(
		    key,
		    event.xbutton.x,
		    event.xbutton.y,
		    (event.xbutton.state&ShiftMask) ? 1 : 0,
		    (event.xbutton.state&ControlMask) ? 1 : 0,
		    (event.xbutton.state&Mod1Mask) ? 1 : 0
		);
		break;
	}
	if(XPending(display)==0) break;
    }
    return 0;
}

static int nil(
Display         *,
XErrorEvent     *
)
{
    return 0;
}

void *eventSocket()
{
    if(display==0)
    {
	XSetErrorHandler(nil);
	XSetIOErrorHandler((XIOErrorHandler)nil);

	const char *displayName= getenv("DISPLAY");
	if(displayName==0) displayName=":0";
	display= XOpenDisplay(displayName);
	if(display==0) system::error(1,"fatal: can't open display %s\n", displayName);

	int screen= DefaultScreen(display);
	Window root= RootWindow(display,screen);
	int rootWidth= XDisplayWidth(display,screen);
	int rootHeight= XDisplayHeight(display,screen);

	int i= 0;
	int attributes[256];
	attributes[i++]= GLX_LEVEL;	attributes[i++]= 0;
	attributes[i++]= GLX_DOUBLEBUFFER;
	attributes[i++]= GLX_RGBA;
	attributes[i++]= GLX_RED_SIZE;	attributes[i++]= 1;
	attributes[i++]= GLX_GREEN_SIZE;attributes[i++]= 1;
	attributes[i++]= GLX_BLUE_SIZE;	attributes[i++]= 1;
	attributes[i++]= GLX_ALPHA_SIZE;attributes[i++]= 0;
	attributes[i++]= GLX_DEPTH_SIZE;attributes[i++]= 4;
	attributes[i++]= (int)None;

	XVisualInfo *visualInfo= glXChooseVisual(display,screen,attributes);
	if(visualInfo==0) system::error(1,"fatal: can't get a decent OpenGL visual\n");

	Colormap colorMap= XCreateColormap(display,root,visualInfo->visual,AllocNone);

	XSetWindowAttributes windowAttributes;
	windowAttributes.colormap=             colorMap;
	windowAttributes.background_pixmap=    None;
	windowAttributes.border_pixel=         0;
	windowAttributes.override_redirect=    0;
	windowAttributes.save_under=           0;

	int width= 720;
	int height= 576;
	window=  XCreateWindow(
	    display,
	    root,
	    (rootWidth-width)/2,
	    (rootHeight-height)/2,
	    width,
	    height,
	    0,
	    visualInfo->depth,
	    InputOutput,
	    visualInfo->visual,
	    CWBackPixmap| CWBorderPixel| CWColormap,
	    &windowAttributes
	);

	XSelectInput(
	    display,
	    window,
	    ButtonMotionMask	|
	    ButtonPressMask	|
	    ButtonReleaseMask	|
	    EnterWindowMask	|
	    ExposureMask	|
	    KeyPressMask	|
	    KeyReleaseMask	|
	    LeaveWindowMask	|
	    PointerMotionMask	|
	    StructureNotifyMask	|
	    VisibilityChangeMask
	);

	XTextProperty tp;
	const char *name= "GPS3D -- OpenGL GPS Viewer";
	XStringListToTextProperty((char**)&name,1,&tp);
	XSetWMProperties(display,window,&tp,&tp,0,0,0,0,0);
	XFree(tp.value);

	GLXContext glxContext= glXCreateContext(display,visualInfo,None,GL_TRUE);
	glXMakeCurrent(display,window,glxContext);

	XMapRaised(display,window);
	XFlush(display);
    }
    return (void*)ConnectionNumber(display);
}

void flush()
{
    glFlush();
    glXSwapBuffers(display,window);
}

#endif // X11

#ifdef macosx

#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <AGL/agl.h>
#include <Carbon/Carbon.h>

enum
{
    kBootStrap= 0,
    kKeyDown,
    kKeyUp,
    kMouseDown,
    kMouseUp,
    kMouseMove,
    kWindowChanged
};

class Msg
{
public:
        int		t;
        int		x;
        int		y;
        int		w;
        int		h;
        int		v;
        int		shift;
        int		control;
        int		alt;
        Msg		*next;
};

static Msg		*queueHead;
static Msg		*queueTail;
static pthread_mutex_t	queueLock;
static int		selectPipe[2];

static void queueMsg(
Msg *msg
)
{
    pthread_mutex_lock(&queueLock);
        if(queueTail) queueTail->next= msg;
        if(queueHead==0) queueHead= msg;
        queueTail= msg;
        msg->next= 0;
        
        int c= 0x12345678;
        write(selectPipe[1],&c,1);
    pthread_mutex_unlock(&queueLock);
}

static WindowRef	window;
static WindowRef	aboutWin;
static AGLContext	glContext;

static int getButtonVal(
EventRef event
)
{
    EventMouseButton whichButton;
    GetEventParameter(
        event,
        kEventParamMouseButton,
        typeMouseButton,
        NULL,
        sizeof(EventMouseButton),
        NULL,
        &whichButton
    );
    
    int v;
         if (whichButton==kEventMouseButtonSecondary)	v=2;
    else if (whichButton==kEventMouseButtonTertiary)	v=1;
    else						v=0;
    return v;
}

static int getKeyVal(
EventRef event
)
{
    UInt32 m;
    GetEventParameter(
        event,
        kEventParamKeyModifiers,
        typeUInt32,
        NULL,
        sizeof(UInt32),
        NULL,
        &m
    );
    
    SInt8 skey;
    GetEventParameter(
        event,
        kEventParamKeyMacCharCodes,
        typeChar,
        NULL,
        sizeof(typeChar),
        NULL,
        &skey
    );
    
    int key= skey;
    if(m&cmdKey)	key+= (97+63);    
    if(m&controlKey)	key+= 96;

         if(key==kEnterCharCode)	key= system::EnterKey;
    else if(key==kHomeCharCode)		key= system::HomeKey;
    else if(key==kTabCharCode)		key= system::TabKey;
    else if(key==kReturnCharCode)	key= system::EnterKey;
    else if(key==kEscapeCharCode)	key= system::EscapeKey;
    
    return key;
}

static void getKeyStates(
Msg		*m,
EventRef	event
)
{
    UInt32 state;
    GetEventParameter(
        event,
        kEventParamKeyModifiers,
        typeUInt32,
        0,
        sizeof(UInt32),
        0,
        &state
    );
    
    m->shift= (state&shiftKey)?1:0;
    m->control= (state&controlKey)?1:0;
    m->alt= (state&optionKey)?1:0;
}

static void syncShape(
int	*x,
int	*y,
int	*w,
int	*h
)
{
    Rect rect;
    GetWindowBounds(
        window,
        kWindowContentRgn,
        &rect
    );

    *x= rect.left;
    *y= rect.bottom;
    *w= rect.right-rect.left;
    *h= rect.bottom-rect.top;
}

static OSStatus aboutHandler(
EventHandlerCallRef	nextHandler,
EventRef		event,
void			*data
)
{
    HideWindow(aboutWin);
    return noErr;
}

static OSStatus eventHandler(
EventHandlerCallRef	nextHandler,
EventRef		event,
void			*data
)
{
    int x,y,w,h;
    syncShape(&x,&y,&w,&h);

    static bool firstTime= true;
    if(firstTime)
    {
        Msg *bootMsg= new Msg;
        bootMsg->t= kBootStrap;
        bootMsg->w= w;
        bootMsg->h= h;
        queueMsg(bootMsg);
        firstTime= false;
    }
    
    Point ePosition;
    UInt32 eKind= GetEventKind(event);
    UInt32 eClass= GetEventClass(event);
    GetEventParameter(
        event,
        kEventParamMouseLocation,
        typeQDPoint,
        NULL,
        sizeof(Point),
        NULL,
        &ePosition
    );

    Msg *m= new Msg;
    m->x= ePosition.h-x;
    m->y= h-(y-ePosition.v);
    
    if(
        eClass==kEventClassMouse	&&
        eKind==kEventMouseUp
    )
    {
        m->t= kMouseUp;
        m->v= getButtonVal(event);
    }
    else if(
        eClass==kEventClassMouse	&&
        eKind==kEventMouseDown
    )
    {
        m->t= kMouseDown;
        m->v= getButtonVal(event);
        getKeyStates(m,event);
    }
    else if(
        eClass==kEventClassMouse	&&
        (
            eKind==kEventMouseMoved	||
            eKind==kEventMouseDragged
        )
    )
    {
        m->t= kMouseMove;
    }
    else if(
        eClass==kEventClassKeyboard	&&
        (
            eKind==kEventRawKeyDown	||
            eKind==kEventRawKeyRepeat
        )
    )
    {
        m->t= kKeyDown;
        m->v= getKeyVal(event);
    }
    else if(
        eClass==kEventClassKeyboard	&&
        eKind==kEventRawKeyUp
    )
    {
        m->t= kKeyUp;
        m->v= getKeyVal(event);
    }
    else if(
        eClass==kEventClassWindow		&&
        (
            eKind==kEventWindowBoundsChanged	||
            eKind==kEventWindowDragCompleted
        )
    )
    {
        m->t=kWindowChanged;
        syncShape(
            &(m->x),
            &(m->y),
            &(m->w),
            &(m->h)
        );
        if(glContext) aglUpdateContext(glContext);
    }
    else if(
        eClass==kEventClassWindow	&&
        eKind==kEventWindowClosed
    )
    {
        system::exit(0);
    }
    else if(
        eClass==kEventClassCommand	&&
        eKind==kEventProcessCommand
    )
    {
        ShowWindow(aboutWin);
        SelectWindow(aboutWin);
    }

    queueMsg(m);
    return eventNotHandledErr;
}

void osxLoop()
{
    IBNibRef nibRef;
    OSStatus err= CreateNibReference(CFSTR("main"),&nibRef);
        if(err!=noErr) system::error(1,"fatal: can't create nib reference.\n");
        
        err = SetMenuBarFromNib(nibRef,CFSTR("MenuBar"));
        if(err!=noErr) system::error(1,"fatal: can't SetMenuBarFromNib.\n");
        
        err= CreateWindowFromNib(nibRef,CFSTR("MainWindow"),&window);
        if(err!=noErr) system::error(1,"fatal: can't CreateWindowFromNib.\n");
        
        err= CreateWindowFromNib(nibRef,CFSTR("AboutWindow"),&aboutWin);
        if(err!=noErr) system::error(1,"fatal: can't CreateWindowFromNib.\n");
    DisposeNibReference(nibRef);
    
    SetPortWindowPort(window);
    
    GLint attr[3];
    attr[0]= AGL_RGBA;
    attr[1]= AGL_DOUBLEBUFFER;
    attr[2]= AGL_NONE; 
    
    AGLPixelFormat format= aglChoosePixelFormat(NULL,0,attr);
    glContext= aglCreateContext(format,NULL);
    aglDestroyPixelFormat(format);
    
    if(glContext==0) system::error(1,"fatal: can't get OpenGL context.\n");
    aglSetDrawable(glContext,GetWindowPort(window));

    EventTypeSpec evList1[]=
    {
        {kEventClassMouse,	kEventMouseDown			},
        {kEventClassMouse,	kEventMouseUp			},
        {kEventClassMouse,	kEventMouseMoved		},
        {kEventClassMouse,	kEventMouseDragged		},
        {kEventClassMouse,	kEventMouseWheelMoved		},
        {kEventClassKeyboard,	kEventRawKeyDown		},
        {kEventClassKeyboard,	kEventRawKeyRepeat		},
        {kEventClassKeyboard,	kEventRawKeyUp			},
        {kEventClassWindow,	kEventWindowActivated		},
        {kEventClassWindow,	kEventWindowClosed		},
        {kEventClassCommand,	kEventProcessCommand		},	 
        {kEventClassAppleEvent,	kEventAppleEvent		},
    };
    InstallApplicationEventHandler(
        NewEventHandlerUPP(eventHandler),
        12,
        evList1,
        window,
        NULL
    );

    EventTypeSpec evList2[]=
    {
        {kEventClassWindow,	kEventWindowBoundsChanged	},
        {kEventClassWindow,	kEventWindowDragCompleted	},
        {kEventClassWindow,	kEventWindowClosed		}
    };
    InstallEventHandler(
        GetWindowEventTarget(window), 
        NewEventHandlerUPP(eventHandler),
        3,
        evList2,
        window,
        NULL
    ); 
      
    EventTypeSpec aboutEv[]=
    {
        {kEventClassWindow,	kEventWindowClose		}
    };
    InstallEventHandler(
        GetWindowEventTarget(aboutWin), 
        NewEventHandlerUPP(aboutHandler),
        1,
        aboutEv,
        0,
        NULL
    );   
            
    BringToFront(window);
    ShowWindow(window);
    SelectWindow(window);
    RunApplicationEventLoop();
}

static void process(
Msg *m
)
{
    switch(m->t)
    {
        case kBootStrap:
        {
            aglSetCurrentContext(glContext);
            onWindowSize(m->w,m->h);
            break;
        }
        case kKeyDown:
        {
	    onKeyDown(m->v,0,0,0,0,0);
            break;
        }
        case kMouseDown:
        {
            onButtonDown(
                m->v,
                m->x,
                m->y,
                m->shift,
                m->control,
                m->alt
            );
            break;
        }
        case kMouseMove:
        {
            onMouseMove(
                m->x,
                m->y
            );
            break;
        }
        case kMouseUp:
        {
            onButtonUp(
                m->v,
                m->x,
                m->y
            );
            break;
        }
        case kWindowChanged:
        {
            onWindowSize(
                m->w,
                m->h
            );
            break;
        }
    }
}

void *eventSocket()
{
    static bool initDone= false;
    if(initDone==false)
    {
	initDone= true;
        pipe(selectPipe);
        fcntl(selectPipe[0],F_SETFL,O_NONBLOCK);
        fcntl(selectPipe[1],F_SETFL,O_NONBLOCK);
    }
    return (void*)(selectPipe[0]);
}

int windowInput()
{
    Msg *m;
    pthread_mutex_lock(&queueLock);
        m= queueHead;
        queueHead= 0;
        queueTail= 0;
        
        while(1)
        {
            char buf[16384];
            int n= read(selectPipe[0],buf,16384);
            if(n<0) break;
        }
    pthread_mutex_unlock(&queueLock);
    
    while(m)
    {
        Msg *next= m->next;
        process(m);
        delete m;
        m= next;
    }

    return 0;
}

void flush()
{
    glFlush();
    aglSwapBuffers(glContext);
}

#endif // macosx
